import numpy as np
import pyqtgraph as pg
import colorcet
from AnyQt.QtCore import QRectF, Qt

from pyqtgraph import LabelItem

import Orange.data
from Orange.widgets.visualize.utils.customizableplot import CommonParameterSetter, Updater
from Orange.widgets.visualize.utils.plotutils import PlotItem, GraphicsView, AxisItem
from Orange.widgets.widget import OWWidget, Msg, Input, Output
from Orange.widgets import gui, settings

from orangewidget.utils import saveplot
from orangewidget.utils.visual_settings_dlg import VisualSettingsDialog
from orangecontrib.spectroscopy.data import getx
from orangecontrib.spectroscopy.widgets.owhyper import ImageColorLegend
from orangecontrib.spectroscopy.widgets.owspectra import InteractiveViewBox
from orangecontrib.spectroscopy.widgets.gui import float_to_str_decimals as strdec, pixel_decimals


# put calculation widgets outside the class for easier reuse
#   without the Orange framework or scripting
def sort_data(data):
    wn = getx(data)
    wn_sorting = np.argsort(wn)
    return data[:, wn_sorting]


# TODO check and use scikit-spectra from
#   https://github.com/hughesadam87/scikit-spectra/tree/master/skspec
#   also verify with corr2D R package  / doi: 10.18637/jss.v090.i03
def calc_cos(table1, table2):
    # TODO make selection in panel for dynamic / static (subtract mean)
    table1 = sort_data(table1)
    table2 = sort_data(table2)

    series1 = table1.X - table1.X.mean()
    series2 = table2.X - table2.X.mean()

    sync = series1.T @ series2 / (len(series1) - 1)

    # Hilbert-Noda transformation matrix
    i, j = np.ogrid[:len(series1), :len(series1)]

    # TODO - We could use the code below after allowing numpy>2.0
    # with np.errstate(divide='ignore'):
    hn = np.where(i != j, 1 / (np.pi * (j - i)), 0)

    # asynchronous correlation
    asyn = series1.T @ hn @ series2 / (len(series1) - 1)

    return sync, asyn, series1, series2, getx(table1), getx(table2)
    # TODO handle non continuous data (after cut widget)


class ParameterSetter(CommonParameterSetter):
    LEFT_AXIS_LABEL, TOP_AXIS_LABEL = "Left axis title", "Top axis title"
    FIGTITLE_LABEL_SIZE = "Title font size"
    PLOT_BOX = "Figure and axes"

    @property
    def title_item(self):
        return self.master.fig_title

    def __init__(self, master):
        super().__init__()
        self.master: OWCos = master

    def update_setters(self):
        self.initial_settings = {
            self.PLOT_BOX: {
                self.TITLE_LABEL: {"Title": ("", "")},
                self.FIGTITLE_LABEL_SIZE: self.FONT_SETTING,
                self.TOP_AXIS_LABEL: {
                    self.TOP_AXIS_LABEL: ("", ""),
                },
                self.LEFT_AXIS_LABEL: {
                    self.LEFT_AXIS_LABEL: ("", ""),
                },
                self.AXIS_TITLE_LABEL: self.FONT_SETTING,
                self.AXIS_TICKS_LABEL: self.FONT_SETTING,
            },
            self.LABELS_BOX: {
                self.FONT_FAMILY_LABEL: self.FONT_FAMILY_SETTING,
            },
        }

        def update_top_axis(**settings):
            top_axis = self.master.top_plot.getAxis("top")
            top_axis.setLabel(settings[self.TOP_AXIS_LABEL])
            top_axis.resizeEvent(None)

        def update_left_axis(**settings):
            left_axis = self.master.left_plot.getAxis("left")
            left_axis.setLabel(settings[self.LEFT_AXIS_LABEL])
            left_axis.resizeEvent(None)

        def update_plot_title_font(**settings):
            font = Updater.change_font(self.title_item.item.font(), settings)
            self.title_item.item.setFont(font)
            self.title_item.item.setPlainText(self.title_item.text)

        def update_figtitle(**settings):
            title = self.master.fig_title
            title.setText(settings[self.TITLE_LABEL])
            title.resizeEvent(None)

        def update_axes_titles(**settings):
            Updater.update_axes_titles_font(self.axis_items, **settings)

        def update_axes_ticks(**settings):
            Updater.update_axes_ticks_font(self.axis_items, **settings)

        def update_font_family(**settings):
            for setter in update_plot_title_font, update_axes_titles, update_axes_ticks:
                setter(**settings)

        self._setters[self.PLOT_BOX] = {
            self.TITLE_LABEL: update_figtitle,
            self.FIGTITLE_LABEL_SIZE: update_plot_title_font,
            self.LEFT_AXIS_LABEL: update_left_axis,
            self.TOP_AXIS_LABEL: update_top_axis,
            self.AXIS_TITLE_LABEL: update_axes_titles,
            self.AXIS_TICKS_LABEL: update_axes_ticks,
        }
        self._setters[self.LABELS_BOX] = {
            self.FONT_FAMILY_LABEL: update_font_family,
        }

    @property
    def axis_items(self):
        return [self.master.left_plot.getAxis("left"),
                self.master.top_plot.getAxis("top"),
                self.master.cbarCOS.axis]


class COS2DViewBox(InteractiveViewBox):
    def autoRange2(self):
        if self is not self.graph.cos2Dplot.vb:
            super().autoRange()
            self.graph.cos2Dplot.vb.autoRange()
        else:
            super().autoRange()

    def suggestPadding(self, axis):
        return 0


class OWCos(OWWidget):
    # Widget's name as displayed in the canvas
    name = "2D Correlation Plot"

    # Short widget description
    description = (
        "Perform 2D correlation analysis with series spectra")

    icon = "icons/2dcos.svg"

    graph_name = "plotview"  # need this to show the save button

    # Define inputs and outputs
    class Inputs:
        data1 = Input("Data 1", Orange.data.Table, default=True)
        data2 = Input("Data 2", Orange.data.Table, default=True)

    class Outputs:
        # TODO implement outputting the matrix
        corr_matrix = Output("2D correlation matrix", Orange.misc.DistMatrix, dynamic=False)

    class Error(OWWidget.Error):
        empty_data = Msg("Input data is empty.")

    settingsHandler = settings.DomainContextHandler()
    selector = settings.Setting(0)
    isonum = settings.Setting(5)
    visual_settings = settings.Setting({}, schema_only=True)

    # autocommit = settings.Setting(True)

    want_main_area = True
    resizing_enabled = True

    class Warning(OWWidget.Warning):
        nodata = Msg("No useful data on input!")

    def __init__(self):
        super().__init__()
        self.parameter_setter = ParameterSetter(self)
        VisualSettingsDialog(
            self, self.parameter_setter.initial_settings
        )

        self.cosmat = None
        self.data1 = None
        self.data2 = None

        # control area
        gui.radioButtons(self.controlArea, self, "selector",
                         btnLabels=("Synchronous", "Asynchronous"), box="Plot type",
                         callback=self.plotCOS)
        self.isocurve_spin = gui.spin(self.controlArea, self, "isonum",
                                      minv=0, maxv=9, step=1,
                                      label="Number of curves", box="Isocurves",
                                      callback=self.plotCOS)
        gui.rubber(self.controlArea)
        self.cursorPos = gui.label(self.controlArea, self, "", box="Crosshair")

        # plotting
        crosspen = pg.mkPen(color=(155, 155, 155), width=0.7)

        self.plotview = GraphicsView()
        self.ci = ci = pg.GraphicsLayout()
        self.plotview.scene().sigMouseMoved.connect(self.mouse_moved_viewhelpers)

        self.plotview.setCentralItem(ci)

        ci.layout.setColumnStretchFactor(0, 1)
        ci.layout.setRowStretchFactor(0, 1)
        ci.layout.setColumnStretchFactor(1, 5)
        ci.layout.setRowStretchFactor(1, 5)

        # image
        self.cos2Dplot = PlotItem(viewBox=COS2DViewBox(self),
                                  axisItems={"left": AxisItem("left"), "bottom": AxisItem("bottom"),
                                             "right": AxisItem("right"), "top": AxisItem("top")})
        self.cos2Dplot.buttonsHidden = True
        ci.addItem(self.cos2Dplot, row=1, col=1)
        self.cos2Dplot.getAxis("left").setStyle(showValues=False)
        self.cos2Dplot.showAxis("top")
        self.cos2Dplot.showAxis("right")
        self.cos2Dplot.getAxis("top").setStyle(showValues=False)
        self.cos2Dplot.getAxis("right").setStyle(showValues=False)
        self.cos2Dplot.getAxis("bottom").setStyle(showValues=False)
        self.cos2Dplot.vb.border = 1
        self.cos2Dplot.vb.setAspectLocked(lock=True, ratio=1)
        self.cos2Dplot.vb.setMouseMode(pg.ViewBox.RectMode)
        # crosshair initialization
        self.vLine = pg.InfiniteLine(angle=90, movable=False, pen=crosspen)
        self.hLine = pg.InfiniteLine(angle=0, movable=False, pen=crosspen)
        self.vLine.setZValue(1000)

        # top spectrum plot
        self.top_plot = PlotItem(viewBox=COS2DViewBox(self),
                                 axisItems={"left": AxisItem("left"), "bottom": AxisItem("bottom"),
                                            "right": AxisItem("right"), "top": AxisItem("top")})
        ci.addItem(self.top_plot, row=0, col=1)
        # visual settings
        self.top_plot.showAxis("right")
        self.top_plot.showAxis("top")
        self.top_plot.getAxis("left").setStyle(showValues=False)
        self.top_plot.getAxis("right").setStyle(showValues=False)
        self.top_plot.getAxis("bottom").setStyle(showValues=False)
        # interactive behavior settings
        self.top_plot.vb.setMouseEnabled(x=True, y=False)
        self.top_plot.enableAutoRange(axis='x')
        self.top_plot.setAutoVisible(x=True)
        self.top_plot.buttonsHidden = True
        self.top_plot.setXLink(self.cos2Dplot)
        # crosshair
        self.top_vLine = pg.InfiniteLine(angle=90, movable=False, pen=crosspen)
        self.top_vLine.setZValue(1000)

        # left spectrum plot
        self.left_plot = PlotItem(viewBox=COS2DViewBox(self),
                                  axisItems={"left": AxisItem("left"), "bottom": AxisItem("bottom"),
                                             "right": AxisItem("right"), "top": AxisItem("top")})
        ci.addItem(self.left_plot, row=1, col=0)
        # visual settings
        self.left_plot.showAxis("right")
        self.left_plot.showAxis("top")
        self.left_plot.getAxis("right").setStyle(showValues=False)
        # self.left_plot.getAxis("right").setPen(color='k')
        self.left_plot.getAxis("top").setStyle(showValues=False)
        self.left_plot.getAxis("bottom").setStyle(showValues=False)
        # self.left_plot.setLabel({"top": "Average spectrum"})
        # interactive behavior settings
        self.left_plot.vb.setMouseEnabled(x=False, y=True)
        self.left_plot.enableAutoRange(axis='y')
        self.left_plot.setAutoVisible(y=True)
        self.left_plot.buttonsHidden = True
        self.left_plot.setYLink(self.cos2Dplot)
        self.left_plot.getViewBox().invertX(True)
        # crosshair
        self.left_hLine = pg.InfiniteLine(angle=0, movable=False, pen=crosspen)
        self.left_hLine.setZValue(1000)

        # colorbar
        self.cbarCOS = ImageColorLegend()
        ci.layout.addItem(self.cbarCOS, 1, 3, 1, 1, alignment=Qt.AlignLeft)

        # moving, resizing and zooming events handling
        self.cos2Dplot.vb.sigRangeChanged.connect(self.update_crosshair)
        self.cos2Dplot.vb.sigResized.connect(self.update_crosshair)

        self.mainArea.layout().addWidget(self.plotview)

        self.important_decimals = 1, 1

        # figure title
        # TODO Orange should implement dark mode change for LabelItem
        self.fig_title = LabelItem(color=(0, 0, 0))
        self.plotview.addItem(self.fig_title)

        # gui.auto_commit(self.controlArea, self, "autocommit", "Apply")

    # TODO - implement the aspect ratio lock for the 2D plot
    #   initialize the widget with the right aspect ratio so that
    #   the pixel is square keep the pixels always square, especially
    #   when changing the widget size
    # TODO - zooming would be better with no scrolling but rather selecting
    #   a range on top or lef or a square on 2D plot
    # TODO - implement rescale Y after zoom
    # TODO make crosshair a black/white double line for better visibility
    # TODO save images with higher resolution by default
    # TODO add ParameterSetter class for using the new fancy GUI
    # TODO optimize initial figure annotations
    # TODO rescale on connect/disconnect

    @Inputs.data1
    def set_data1(self, dataset):
        self.data1 = dataset

    @Inputs.data2
    def set_data2(self, dataset):
        self.data2 = dataset

    def set_visual_settings(self, key, value):
        self.parameter_setter.set_parameter(key, value)
        self.visual_settings[key] = value

    def handleNewSignals(self):
        self.cos2Dplot.clear()
        self.Error.empty_data.clear()
        self.cosmat = None

        d1 = self.data1
        d2 = self.data2
        if d1 is None:
            d1, d2 = d2, d1

        if not d2:
            d2 = d1

        if (not d1 or not d2 or not d1.X.size or not d2.X.size or
                np.all(np.isnan(d1.X)) or np.all(np.isnan(d2.X))):
            self.Error.empty_data()
        else:
            self.cosmat = calc_cos(d1, d2)

        self.commit()

    def mouse_moved_viewhelpers(self, pos):
        if self.cos2Dplot.sceneBoundingRect().contains(pos):
            mousePoint = self.cos2Dplot.vb.mapSceneToView(pos)
            self.setCrosshairPos(mousePoint.x(), mousePoint.y())

        if self.left_plot.sceneBoundingRect().contains(pos):
            mousePoint = self.left_plot.vb.mapSceneToView(pos)
            self.setCrosshairPos(None, mousePoint.y())

        if self.top_plot.sceneBoundingRect().contains(pos):
            mousePoint = self.top_plot.vb.mapSceneToView(pos)
            self.setCrosshairPos(mousePoint.x(), None)

    def setCrosshairPos(self, x, y):
        if x is not None:
            self.vLine.setPos(x)
            self.top_vLine.setPos(x)

        if y is not None:
            self.left_hLine.setPos(y)
            self.hLine.setPos(y)

        x = 'None' if x is None else strdec(x, self.important_decimals[0])
        y = 'None' if y is None else strdec(y, self.important_decimals[1])

        self.cursorPos.setText(f"{x}, {y}")

    def update_crosshair(self):
        self.important_decimals = pixel_decimals(self.cos2Dplot.vb)

    def plot_type_change(self):
        self.commit()

    def plotCOS(self):
        self.cos2Dplot.clear()
        self.left_plot.clear()
        self.top_plot.clear()

        p = pg.mkPen('r', width=2)

        if self.cosmat is not None:
            cosmat = self.cosmat[self.selector]
            topSP = self.cosmat[2]
            leftSP = self.cosmat[3]
            topSPwn = self.cosmat[4]
            leftSPwn = self.cosmat[5]

            COSimage = pg.ImageItem(image=cosmat)
            COSimage.setLevels([-1 * np.absolute(cosmat).max(), np.absolute(cosmat).max()])
            COSimage.setLookupTable(np.array(colorcet.diverging_bwr_40_95_c42) * 255)

            self.cos2Dplot.addItem(COSimage)
            self.cos2Dplot.addItem(self.vLine, ignoreBounds=True)
            self.cos2Dplot.addItem(self.hLine, ignoreBounds=True)

            # adding iso curves
            # TODO make this multi threaded because big images slow the widget down very much
            #   good example in owtranspose
            level_max = np.max([np.abs(np.min(cosmat)), np.abs(np.max(cosmat))])
            levels = np.linspace(start=0, stop=level_max, num=self.isonum + 2)
            iso_pen_pos = pg.mkPen(color=(50, 50, 50), width=1)
            iso_pen_neg = pg.mkPen(color=(50, 50, 50), width=1, style=Qt.DashLine)
            for level in levels[1:]:
                ic = pg.IsocurveItem(data=cosmat, level=level, pen=iso_pen_pos)
                ic.setParentItem(COSimage)
                ic = pg.IsocurveItem(data=cosmat, level=-level, pen=iso_pen_neg)
                ic.setParentItem(COSimage)

            COSimage.setRect(QRectF(topSPwn.min(),
                                    leftSPwn.min(),
                                    (topSPwn.max() - topSPwn.min()),
                                    (leftSPwn.max() - leftSPwn.min())))

            self.cbarCOS.set_range(-1 * np.nanmax(np.absolute(cosmat)), np.nanmax(np.absolute(cosmat)))
            self.cbarCOS.set_colors(np.array(colorcet.diverging_bwr_40_95_c42) * 255)

            left_indices = np.linspace(0, len(leftSP)-1, min(100, len(leftSP)), dtype=int)

            for s in leftSP[left_indices]:
                pt = pg.PlotCurveItem(s, leftSPwn, pen=pg.mkPen(color=(50, 50, 50), width=0.5))
                self.left_plot.addItem(pt, ignoreBounds=True)

            self.left_plot.plot(leftSP.mean(axis=0), leftSPwn, pen=p)
            self.left_plot.addItem(self.left_hLine)

            self.left_plot.setXRange(np.min(leftSP), np.max(leftSP))

            top_indices = np.linspace(0, len(topSP)-1, min(100, len(topSP)), dtype=int)

            for s in topSP[top_indices]:
                pt = pg.PlotCurveItem(topSPwn, s, pen=pg.mkPen(color=(50, 50, 50), width=0.5))
                self.top_plot.addItem(pt, ignoreBounds=True)

            self.top_plot.plot(topSPwn, topSP.mean(axis=0), pen=p)
            self.top_plot.addItem(self.top_vLine)

            self.top_plot.setYRange(np.min(topSP), np.max(topSP))

    def hideitems_for_saving(self):
        self.hLine.hide()
        self.vLine.hide()
        self.left_hLine.hide()
        self.top_vLine.hide()

    def showitems(self):
        self.hLine.show()
        self.vLine.show()
        self.left_hLine.show()
        self.top_vLine.show()

    def save_graph(self):
        try:
            self.hideitems_for_saving()
            saveplot.save_plot(self.plotview, self.graph_writers)
        finally:
            self.showitems()

    def commit(self):
        self.plotCOS()


if __name__ == "__main__":  # pragma: no cover
    # pylint: disable=ungrouped-imports
    from Orange.widgets.utils.widgetpreview import WidgetPreview

    data1 = Orange.data.Table("collagen")[100:230]
    data2 = Orange.data.Table("collagen")[600:730]
    WidgetPreview(OWCos).run(set_data1=data1, set_data2=data2)



import random
import unittest

import numpy as np

import Orange
from Orange.classification import RandomForestLearner
from Orange.data import Table
from Orange.evaluation import TestOnTestData, AUC

from orangecontrib.spectroscopy.data import getx
from orangecontrib.spectroscopy.preprocess import Absorbance, Transmittance, \
    Integrate, Interpolate, SavitzkyGolayFiltering, \
    GaussianSmoothing, PCADenoising, RubberbandBaseline, \
    Normalize, LinearBaseline, ShiftAndScale, MissingReferenceException, \
    WrongReferenceException, NormalizeReference, \
    PreprocessException, NormalizePhaseReference, SpSubtract, MNFDenoising
from orangecontrib.spectroscopy.preprocess.utils import replacex
from orangecontrib.spectroscopy.tests.test_conversion import separate_learn_test, slightly_change_wavenumbers, odd_attr
from orangecontrib.spectroscopy.tests.util import smaller_data


COLLAGEN = Orange.data.Table("collagen")
SMALL_COLLAGEN = smaller_data(COLLAGEN, 2, 2)
SMALLER_COLLAGEN = smaller_data(COLLAGEN[195:621], 40, 4)  # only glycogen and lipids


def add_zeros(data):
    """ Every 5th value is zero """
    s = data.copy()
    with s.unlocked():
        s[:, ::5] = 0
    return s


def make_edges_nan(data):
    s = data.copy()
    with s.unlocked():
        s[:, 0:3] = np.nan
        s[:, s.X.shape[1]-3:] = np.nan
    return s


def make_middle_nan(data):
    """ Four middle values are NaN """
    s = data.copy()
    half = s.X.shape[1]//2
    with s.unlocked():
        s[:, half-2:half+2] = np.nan
    return s


def shuffle_attr(data):
    natts = list(data.domain.attributes)
    random.Random(0).shuffle(natts)
    ndomain = Orange.data.Domain(natts, data.domain.class_vars,
                                 metas=data.domain.metas)
    return data.transform(ndomain)


def reverse_attr(data):
    natts = reversed(data.domain.attributes)
    ndomain = Orange.data.Domain(natts, data.domain.class_vars,
                                 metas=data.domain.metas)
    return data.transform(ndomain)


def add_edge_case_data_parameter(class_, data_arg_name, data_to_modify, *args, **kwargs):
    modified = [data_to_modify,
                shuffle_attr(data_to_modify),
                make_edges_nan(data_to_modify),
                shuffle_attr(make_edges_nan(data_to_modify)),
                make_middle_nan(data_to_modify),
                add_zeros(data_to_modify),
                ]
    for i, d in enumerate(modified):
        kwargs[data_arg_name] = d
        p = class_(*args, **kwargs)
        # 5 is add_zeros
        if i == 5:
            p.skip_add_zeros = True
        yield p


class TestConversionMixin:

    def test_slightly_different_domain(self):
        """ If test data has a slightly different domain then (with interpolation)
        we should obtain a similar classification score. """
        learner = RandomForestLearner(random_state=42)

        for proc in self.preprocessors:
            if hasattr(proc, "skip_add_zeros"):
                continue
            with self.subTest(proc):
                # LR that can not handle unknown values
                train, test = separate_learn_test(self.data)
                train1 = proc(train)
                aucorig = AUC(TestOnTestData()(train1, test, [learner]))
                test = slightly_change_wavenumbers(test, 0.00001)
                test = odd_attr(test)
                # a subset of points for training so that all test sets points
                # are within the train set points, which gives no unknowns
                train = Interpolate(points=getx(train)[1:-3])(train)  # interpolatable train
                train = proc(train)
                # explicit domain conversion test to catch exceptions that would
                # otherwise be silently handled in TestOnTestData
                _ = test.transform(train.domain)
                aucnow = AUC(TestOnTestData()(train, test, [learner]))
                self.assertAlmostEqual(aucnow, aucorig, delta=0.03, msg="Preprocessor " + str(proc))
                test = Interpolate(points=getx(test) - 1.)(test)  # also do a shift
                _ = test.transform(train.domain)  # explicit call again
                aucnow = AUC(TestOnTestData()(train, test, [learner]))
                # the difference should be slight
                self.assertAlmostEqual(aucnow, aucorig, delta=0.05, msg="Preprocessor " + str(proc))


class TestConversionIndpSamplesMixin(TestConversionMixin):
    """
    Testing mixin for preprocessors that work per sample and should
    return the same result for a sample independent of the other samples
    """

    def test_whole_and_train_separate(self):
        """ Applying a preprocessor before spliting data into train and test
        and applying is just on train data should yield the same transformation of
        the test data. """
        for proc in self.preprocessors:
            with self.subTest(proc):
                data = self.data
                _, test1 = separate_learn_test(proc(data))
                train, test = separate_learn_test(data)
                train = proc(train)
                test_transformed = test.transform(train.domain)
                np.testing.assert_almost_equal(test_transformed.X, test1.X,
                                               err_msg="Preprocessor " + str(proc))


class TestStrangeDataMixin:

    def test_no_samples(self):
        """ Preprocessors should not crash when there are no input samples. """
        data = self.data[:0]
        for proc in self.preprocessors:
            with self.subTest(proc):
                _ = proc(data)

    def test_no_attributes(self):
        """ Preprocessors should not crash when samples have no attributes. """
        data = self.data
        data = data.transform(Orange.data.Domain([],
                                                 class_vars=data.domain.class_vars,
                                                 metas=data.domain.metas))
        for proc in self.preprocessors:
            with self.subTest(proc):
                _ = proc(data)

    def test_all_nans(self):
        """ Preprocessors should not crash when there are all-nan samples. """
        for proc in self.preprocessors:
            with self.subTest(proc):
                data = self.data.copy()
                with data.unlocked():
                    data.X[0, :] = np.nan
                try:
                    _ = proc(data)
                except PreprocessException:
                    continue  # allow explicit preprocessor exception

    def test_unordered_features(self):
        for proc in self.preprocessors:
            with self.subTest(proc):
                data = self.data
                data_reversed = reverse_attr(data)
                data_shuffle = shuffle_attr(data)
                pdata = proc(data)
                X = pdata.X[:, np.argsort(getx(pdata))]
                pdata_reversed = proc(data_reversed)
                X_reversed = pdata_reversed.X[:, np.argsort(getx(pdata_reversed))]
                np.testing.assert_almost_equal(X, X_reversed, err_msg="Preprocessor " + str(proc))
                pdata_shuffle = proc(data_shuffle)
                X_shuffle = pdata_shuffle.X[:, np.argsort(getx(pdata_shuffle))]
                np.testing.assert_almost_equal(X, X_shuffle, err_msg="Preprocessor " + str(proc))

    def test_unknown_no_propagate(self):
        for proc in self.preprocessors:
            with self.subTest(proc):
                data = self.data.copy()
                # one unknown in line
                with data.unlocked():
                    for i in range(min(len(data), len(data.domain.attributes))):
                        data.X[i, i] = np.nan

                if hasattr(proc, "skip_add_zeros"):
                    continue
                pdata = proc(data)
                sumnans = np.sum(np.isnan(pdata.X), axis=1)
                self.assertFalse(np.any(sumnans > 1), msg="Preprocessor " + str(proc))

    def test_no_infs(self):
        """ Preprocessors should not return (-)inf """
        for proc in self.preprocessors:
            with self.subTest(proc):
                data = self.data.copy()
                # add some zeros to the dataset
                with data.unlocked():
                    for i in range(min(len(data), len(data.domain.attributes))):
                        data.X[i, i] = 0
                    data.X[0, :] = 0
                    data.X[:, 0] = 0
                try:
                    pdata = proc(data)
                except PreprocessException:
                    continue  # allow explicit preprocessor exception
                anyinfs = np.any(np.isinf(pdata.X))
                self.assertFalse(anyinfs, msg="Preprocessor " + str(proc))


class TestCommonMixin(TestStrangeDataMixin, TestConversionMixin):
    pass


class TestCommonIndpSamplesMixin(TestStrangeDataMixin, TestConversionIndpSamplesMixin):
    pass


class TestSpSubtract(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors = list(add_edge_case_data_parameter(
        SpSubtract, "reference", SMALLER_COLLAGEN[:1], amount=0.1))
    data = SMALLER_COLLAGEN

    def test_simple(self):
        data = Table.from_numpy(None, [[1.0, 2.0, 3.0, 4.0]])
        reference = Table.from_numpy(None, [[1.0, 2.0, 3.0, 4.0]])
        f = SpSubtract(reference, amount=2)
        fdata = f(data)
        np.testing.assert_almost_equal(fdata.X, [[-1.0, -2.0, -3.0, -4.0]])


class TestTransmittance(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors =  [Transmittance()] + \
                      list(add_edge_case_data_parameter(
                          Transmittance, "reference", SMALLER_COLLAGEN[0:1]))
    data = SMALLER_COLLAGEN

    def test_domain_conversion(self):
        """Test whether a domain can be used for conversion."""
        data = self.data
        transmittance = Transmittance()(data)
        nt = Orange.data.Table.from_table(transmittance.domain, data)
        self.assertEqual(transmittance.domain, nt.domain)
        np.testing.assert_equal(transmittance.X, nt.X)
        np.testing.assert_equal(transmittance.Y, nt.Y)

    def test_roundtrip(self):
        """Test AB -> TR -> AB calculation"""
        data = self.data
        calcdata = Absorbance()(Transmittance()(data))
        np.testing.assert_allclose(data.X, calcdata.X)

    def test_eq(self):
        data = self.data
        t1 = Transmittance()(data)
        t2 = Transmittance()(data)
        self.assertEqual(t1.domain, t2.domain)
        data2 = Table.from_numpy(None, [[1.0, 2.0, 3.0, 4.0]])
        t3 = Transmittance()(data2)
        self.assertNotEqual(t1.domain, t3.domain)
        t4 = Transmittance(reference=data2)(data)
        self.assertNotEqual(t1.domain, t4.domain)
        t5 = Transmittance(reference=data2)(data[:1])
        self.assertGreater(len(t4), len(t5))
        self.assertEqual(t4.domain, t5.domain)
        a = Absorbance()(data)
        self.assertNotEqual(a.domain, t1.domain)


class TestAbsorbance(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors =  [Absorbance()] + \
                      list(add_edge_case_data_parameter(
                          Absorbance, "reference", SMALLER_COLLAGEN[0:1]))
    data = SMALLER_COLLAGEN


    def test_domain_conversion(self):
        """Test whether a domain can be used for conversion."""
        data = Transmittance()(self.data)
        absorbance = Absorbance()(data)
        nt = Orange.data.Table.from_table(absorbance.domain, data)
        self.assertEqual(absorbance.domain, nt.domain)
        np.testing.assert_equal(absorbance.X, nt.X)
        np.testing.assert_equal(absorbance.Y, nt.Y)

    def test_roundtrip(self):
        """Test TR -> AB -> TR calculation"""
        # actually AB -> TR -> AB -> TR
        data = Transmittance()(self.data)
        calcdata = Transmittance()(Absorbance()(data))
        np.testing.assert_allclose(data.X, calcdata.X)

    def test_eq(self):
        data = self.data
        t1 = Absorbance()(data)
        t2 = Absorbance()(data)
        self.assertEqual(t1.domain, t2.domain)
        data2 = Table.from_numpy(None, [[1.0, 2.0, 3.0, 4.0]])
        t3 = Absorbance()(data2)
        self.assertNotEqual(t1.domain, t3.domain)
        t4 = Absorbance(reference=data2)(data)
        self.assertNotEqual(t1.domain, t4.domain)
        t5 = Absorbance(reference=data2)(data[:1])
        self.assertGreater(len(t4), len(t5))
        self.assertEqual(t4.domain, t5.domain)


class TestSavitzkyGolay(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors =  [SavitzkyGolayFiltering(window=9, polyorder=2, deriv=2)]
    data = SMALL_COLLAGEN

    def test_simple(self):
        data = Orange.data.Table("iris")
        f = SavitzkyGolayFiltering()
        data = data[:1]
        fdata = f(data)
        np.testing.assert_almost_equal(fdata.X,
                                       [[4.86857143, 3.47428571, 1.49428571, 0.32857143]])

    def test_eq(self):
        data = Table.from_numpy(None, [[2, 1, 2, 2, 3]])
        p1 = SavitzkyGolayFiltering(window=5, polyorder=2, deriv=0)(data)
        p2 = SavitzkyGolayFiltering(window=5, polyorder=2, deriv=1)(data)
        p3 = SavitzkyGolayFiltering(window=5, polyorder=3, deriv=0)(data)
        p4 = SavitzkyGolayFiltering(window=7, polyorder=2, deriv=0)(data)
        self.assertNotEqual(p1.domain, p2.domain)
        self.assertNotEqual(p1.domain, p3.domain)
        self.assertNotEqual(p1.domain, p4.domain)

        s1 = SavitzkyGolayFiltering(window=5, polyorder=2, deriv=0)(data)
        self.assertEqual(p1.domain, s1.domain)

        # even if the data set is different features should be the same
        data2 = Table.from_numpy(None, [[2, 1, 3, 4, 3]])
        s2 = SavitzkyGolayFiltering(window=5, polyorder=2, deriv=0)(data2)
        self.assertEqual(p1.domain, s2.domain)


class TestGaussian(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors = [GaussianSmoothing(sd=3.)]
    data = SMALL_COLLAGEN

    def test_simple(self):
        data = Orange.data.Table("iris")
        f = GaussianSmoothing(sd=1.)
        data = data[:1]
        fdata = f(data)
        np.testing.assert_almost_equal(fdata.X,
                                       [[4.4907066, 3.2794677, 1.7641664, 0.6909083]])


class TestRubberbandBaseline(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors =  [RubberbandBaseline()]
    data = SMALLER_COLLAGEN

    def test_whole(self):
        """ Every point belongs in the convex region. """
        data = Table.from_numpy(None, [[2, 1, 2]])
        i = RubberbandBaseline()(data)
        np.testing.assert_equal(i.X, 0)
        data = Table.from_numpy(None, [[1, 2, 1]])
        i = RubberbandBaseline(peak_dir=RubberbandBaseline.PeakNegative)(data)
        np.testing.assert_equal(i.X, 0)

    def test_simple(self):
        """ Just one point is not in the convex region. """
        data = Table.from_numpy(None, [[1, 2, 1, 1]])
        i = RubberbandBaseline()(data)
        np.testing.assert_equal(i.X, [[0, 1, 0, 0]])
        data = Table.from_numpy(None, [[1, 2, 1, 1]])
        i = RubberbandBaseline(peak_dir=RubberbandBaseline.PeakNegative)(data)
        np.testing.assert_equal(i.X, [[0, 0, -0.5, 0]])


class TestLinearBaseline(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors =  [LinearBaseline()]
    data = SMALL_COLLAGEN

    def test_whole(self):
        data = Table.from_numpy(None, [[1, 5, 1]])
        i = LinearBaseline()(data)
        np.testing.assert_equal(i.X, [[0, 4, 0]])

        data = Table.from_numpy(None, [[4, 1, 2, 4]])
        i = LinearBaseline(peak_dir=LinearBaseline.PeakNegative)(data)
        np.testing.assert_equal(i.X, [[0, -3, -2, 0]])

    def test_edgepoints(self):
        data = Table.from_numpy(None, [[1, 5, 1]])
        i = LinearBaseline(zero_points=[0, 2])(data)
        np.testing.assert_equal(i.X, [[0, 4, 0]])

    def test_edgepoints_extrapolate(self):
        data = Table.from_numpy(None, [[1, 5, 1]])
        i = LinearBaseline(zero_points=[0, 1])(data)
        np.testing.assert_equal(i.X, [[0, 0, -8]])

    def test_3points(self):
        data = Table.from_numpy(None, [[1, 5, 1, 5]])
        i = LinearBaseline(zero_points=[0, 1, 3])(data)
        np.testing.assert_equal(i.X, [[0, 0, -4, 0]])

    def test_edgepoints_out_of_data(self):
        data = Table.from_numpy(None, [[1, 5, 1]])
        i = LinearBaseline(zero_points=[0, 2.000000001])(data)
        np.testing.assert_almost_equal(i.X, [[0, 4, 0]])


class TestNormalize(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors = [Normalize(method=Normalize.Vector),
                     Normalize(method=Normalize.Area,
                               int_method=Integrate.PeakMax, lower=0, upper=10000),
                     Normalize(method=Normalize.MinMax)]

    data = SMALL_COLLAGEN

    def test_vector_norm(self):
        data = Table.from_numpy(None, [[2, 1, 2, 2, 3]])
        p = Normalize(method=Normalize.Vector)(data)
        q = data.X / np.sqrt((data.X * data.X).sum(axis=1))
        np.testing.assert_equal(p.X, q)
        p = Normalize(method=Normalize.Vector, lower=0, upper=4)(data)
        np.testing.assert_equal(p.X, q)
        p = Normalize(method=Normalize.Vector, lower=0, upper=2)(data)
        np.testing.assert_equal(p.X, q)

    def test_vector_norm_nan_correction(self):
        # even though some values are unknown the other values
        # should be normalized to the same results
        data = Table.from_numpy(None, [[2, 2, 2, 2]])
        p = Normalize(method=Normalize.Vector)(data)
        self.assertAlmostEqual(p.X[0, 0], 0.5)
        # unknown in between that can be interpolated does not change results
        with data.unlocked():
            data.X[0, 2] = float("nan")
        p = Normalize(method=Normalize.Vector)(data)
        self.assertAlmostEqual(p.X[0, 0], 0.5)
        self.assertTrue(np.isnan(p.X[0, 2]))
        # unknowns at the edges do not get interpolated
        with data.unlocked():
            data.X[0, 3] = float("nan")
        p = Normalize(method=Normalize.Vector)(data)
        self.assertAlmostEqual(p.X[0, 0], 2**0.5/2)
        self.assertTrue(np.all(np.isnan(p.X[0, 2:])))

    def test_area_norm(self):
        data = Table.from_numpy(None, [[2, 1, 2, 2, 3]])
        p = Normalize(method=Normalize.Area, int_method=Integrate.PeakMax, lower=0, upper=4)(data)
        np.testing.assert_equal(p.X, data.X / 3)
        p = Normalize(method=Normalize.Area, int_method=Integrate.Simple, lower=0, upper=4)(data)
        np.testing.assert_equal(p.X, data.X / 7.5)
        p = Normalize(method=Normalize.Area, int_method=Integrate.Simple, lower=0, upper=2)(data)
        q = Integrate(methods=Integrate.Simple, limits=[[0, 2]])(p)
        np.testing.assert_equal(q.X, np.ones_like(q.X))

    def test_attribute_norm(self):
        data = Table.from_numpy(None, [[2, 1, 2, 2, 3]])
        ndom = Orange.data.Domain(data.domain.attributes, data.domain.class_vars,
                                  metas=[Orange.data.ContinuousVariable("f")])
        data = data.transform(ndom)
        with data.unlocked(data.metas):
            data[0]["f"] = 2
        p = Normalize(method=Normalize.Attribute, attr=data.domain.metas[0])(data)
        np.testing.assert_equal(p.X, data.X / 2)
        p = Normalize(method=Normalize.Attribute, attr=data.domain.metas[0],
                      lower=0, upper=4)(data)
        np.testing.assert_equal(p.X, data.X / 2)
        p = Normalize(method=Normalize.Attribute, attr=data.domain.metas[0],
                      lower=2, upper=4)(data)
        np.testing.assert_equal(p.X, data.X / 2)

    def test_attribute_norm_unknown(self):
        data = Table.from_numpy(None, X=[[2, 1, 2, 2, 3]], metas=[[2]])
        p = Normalize(method=Normalize.Attribute, attr="unknown")(data)
        self.assertTrue(np.all(np.isnan(p.X)))

    def test_minmax_norm(self):
        data = Table.from_numpy(None, [[2, 1, 2, 2, 3]])
        p = Normalize(method=Normalize.MinMax)(data)
        q = (data.X) / (3 - 1)
        np.testing.assert_equal(p.X, q)
        p = Normalize(method=Normalize.MinMax, lower=0, upper=4)(data)
        np.testing.assert_equal(p.X, q)
        p = Normalize(method=Normalize.MinMax, lower=0, upper=2)(data)
        np.testing.assert_equal(p.X, q)

    def test_SNV_norm(self):
        data = Table.from_numpy(None, [[2, 1, 2, 2, 3]])
        p = Normalize(method=Normalize.SNV)(data)
        q = (data.X - 2) / 0.6324555320336759
        np.testing.assert_equal(p.X, q)
        p = Normalize(method=Normalize.SNV, lower=0, upper=4)(data)
        np.testing.assert_equal(p.X, q)
        p = Normalize(method=Normalize.SNV, lower=0, upper=2)(data)
        np.testing.assert_equal(p.X, q)

    def test_eq(self):
        data = Table.from_numpy(None, [[2, 1, 2, 2, 3]])
        p1 = Normalize(method=Normalize.MinMax)(data)
        p2 = Normalize(method=Normalize.SNV)(data)
        p3 = Normalize(method=Normalize.MinMax)(data)
        self.assertNotEqual(p1.domain, p2.domain)
        self.assertEqual(p1.domain, p3.domain)

        p1 = Normalize(method=Normalize.Area, int_method=Integrate.PeakMax,
                       lower=0, upper=4)(data)
        p2 = Normalize(method=Normalize.Area, int_method=Integrate.Baseline,
                       lower=0, upper=4)(data)
        p3 = Normalize(method=Normalize.Area, int_method=Integrate.PeakMax,
                       lower=1, upper=4)(data)
        p4 = Normalize(method=Normalize.Area, int_method=Integrate.PeakMax,
                       lower=0, upper=4)(data)
        self.assertNotEqual(p1.domain, p2.domain)
        self.assertNotEqual(p1.domain, p3.domain)
        self.assertEqual(p1.domain, p4.domain)


class TestNormalizeReference(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors = (list(add_edge_case_data_parameter(NormalizeReference,
                                                      "reference", SMALLER_COLLAGEN[:1])) +
                     list(add_edge_case_data_parameter(NormalizePhaseReference,
                                                      "reference", SMALLER_COLLAGEN[:1])))
    data = SMALLER_COLLAGEN

    def test_reference(self):
        data = Table.from_numpy(None, [[2, 1, 3], [4, 2, 6]])
        reference = data[:1]
        p = NormalizeReference(reference=reference)(data)
        np.testing.assert_almost_equal(p, [[1, 1, 1], [2, 2, 2]])
        s = NormalizePhaseReference(reference=reference)(data)
        np.testing.assert_almost_equal(s, [[0, 0, 0], [2, 1, 3]])

    def test_reference_exceptions(self):
        with self.assertRaises(MissingReferenceException):
            NormalizeReference(reference=None)
        with self.assertRaises(WrongReferenceException):
            NormalizeReference(reference=Table.from_numpy(None, [[2], [6]]))


class TestPCADenoising(unittest.TestCase, TestCommonMixin):

    preprocessors = [PCADenoising(components=2)]
    data = SMALLER_COLLAGEN

    def test_no_samples(self):
        data = self.data
        proc = PCADenoising()
        d1 = proc(data[:0])
        newdata = data.transform(d1.domain)
        np.testing.assert_equal(newdata.X, np.nan)

    def test_iris(self):
        data = Orange.data.Table("iris")
        proc = PCADenoising(components=2)
        d1 = proc(data)
        newdata = data.transform(d1.domain)
        differences = newdata.X - data.X
        self.assertTrue(np.all(np.abs(differences) < 0.6))
        # pin some values to detect changes in the PCA implementation
        # (for example normalization)
        np.testing.assert_almost_equal(newdata.X[:2],
                                       [[5.08718247, 3.51315614, 1.40204280, 0.21105556],
                                        [4.75015528, 3.15366444, 1.46254138, 0.23693223]])

    def test_selected_components(self):
        data = Orange.data.Table("iris")
        proc = PCADenoising(components=[0, 1])
        d1 = proc(data)
        proc2 = PCADenoising(components=2)
        d2 = proc2(data)
        np.testing.assert_equal(d1.X, d2.X)


class TestMNFDenoising(unittest.TestCase, TestCommonMixin):

    preprocessors = [MNFDenoising(components=2)]
    data = SMALL_COLLAGEN

    def test_no_samples(self):
        data = Orange.data.Table("iris")
        proc = MNFDenoising()
        d1 = proc(data[:0])
        newdata = data.transform(d1.domain)
        np.testing.assert_equal(newdata.X, np.nan)

    def test_iris(self):
        data = Orange.data.Table("iris")
        proc = MNFDenoising(components=2)
        d1 = proc(data)
        newdata = data.transform(d1.domain)
        differences = newdata.X - data.X
        self.assertTrue(np.all(np.abs(differences) < 0.6))
        # pin some values to detect changes in the PCA implementation
        # (for example normalization)
        np.testing.assert_almost_equal(newdata.X[:2],
                                       [[5.1084779, 3.4893387, 1.4068703, 0.1887913],
                                        [4.7484942, 3.1913347, 1.427665, 0.2304239]])

    def test_slightly_different_domain(self):
        # test is disabled because this method is too sensitive to small input changes
        pass


class TestShiftAndScale(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors = [ShiftAndScale(1, 2)]
    data = SMALL_COLLAGEN

    def test_simple(self):
        data = Table.from_numpy(None, [[1.0, 2.0, 3.0, 4.0]])
        f = ShiftAndScale(offset=1.1, scale=2.)
        fdata = f(data)
        np.testing.assert_almost_equal(fdata.X,
                                       [[3.1, 5.1, 7.1, 9.1]])


class TestUtils(unittest.TestCase):

    def test_replacex(self):
        data = Table.from_numpy(None, [[1.0, 2.0, 3.0, 4.0]])
        self.assertEqual(list(getx(data)), [0, 1, 2, 3])
        dr = replacex(data, ["a", 1, 2, 3])
        self.assertEqual([a.name for a in dr.domain.attributes],
                         ["a", "1", "2", "3"])
        dr = replacex(data, np.array([0.5, 1, 2, 3]))
        self.assertEqual(list(getx(dr)), [0.5, 1, 2, 3])
        np.testing.assert_equal(data.X, dr.X)

    def test_replacex_transforms(self):
        data = Table.from_numpy(None, [[1.0, 2.0, 3.0, 4.0]])
        dr = replacex(data, np.linspace(5, 8, 4))
        self.assertEqual(list(getx(dr)), [5, 6, 7, 8])
        np.testing.assert_equal(data.X, dr.X)
        dt = data.transform(dr.domain)
        np.testing.assert_equal(data.X, dt.X)

    def test_replacex_invalid(self):
        data = Table.from_numpy(None, [[1.0, 2.0, 3.0, 4.0]])
        with self.assertRaises(AssertionError):
            replacex(data, [1, 2, 3])



import warnings
from typing import List

import h5py
import numpy as np
from Orange.data import FileFormat, ContinuousVariable

from orangecontrib.spectroscopy.io.util import SpectralFileFormat, _spectra_from_image


class HDF5Reader_SGM(FileFormat, SpectralFileFormat):
    """ A very case specific reader for interpolated hyperspectral mapping HDF5
    files from the SGM beamline at the CLS"""
    EXTENSIONS = ('.h5',)
    DESCRIPTION = 'HDF5 file @SGM/CLS'

    @property
    def sheets(self) -> List:
        sheets = ["All"]
        with h5py.File(self.filename, 'r') as h5:
            NXentries = [str(x) for x in h5['/'].keys() if 'NXentry' in str(h5[x].attrs.get('NX_class'))]
            NXdata = [entry + "/" + str(x) for entry in NXentries for x in h5['/' + entry].keys()
                      if 'NXdata' in str(h5[entry + "/" + x].attrs.get('NX_class'))]
            for d in NXdata:
                sheets.extend([k for k, v in h5[d].items() if len(v.shape) == 3 and v.shape[-1] > 1])
        return sheets

    def read_spectra(self):
        if self.sheet is None:
            self.sheet = self.sheets[0]
        with h5py.File(self.filename, 'r') as h5:
            NXentries = [str(x) for x in h5['/'].keys() if 'NXentry' in str(h5[x].attrs.get('NX_class'))]
            NXdata = [entry + "/" + str(x) for entry in NXentries for x in h5['/' + entry].keys()
                      if 'NXdata' in str(h5[entry + "/" + x].attrs.get('NX_class'))]
            axes = [[str(nm) for nm in h5[nxdata].keys() for s in h5[nxdata].attrs.get('axes') if str(s) in str(nm) or
                     str(nm) in str(s)] for nxdata in NXdata]
            indep_shape = [v.shape for i, d in enumerate(NXdata) for k, v in h5[d].items() if k in axes[i][0]]
            data = [
                {k: np.squeeze(v[()]) for k, v in h5[d].items() if v.shape[0] == indep_shape[i][0] and k not in axes[i]}
                for i, d in
                enumerate(NXdata)]

            features_entries = []
            X_entries = []
            meta_table_entries = []
            for i, d in enumerate(NXdata):
                d = NXdata[i]

                if len(axes[i]) == 1:
                    warnings.warn(f"1D datasets not yet implemented: {d} not loaded.")
                x_locs = h5[d][axes[i][0]]
                y_locs = h5[d][axes[i][1]]
                en = h5[d]['en']
                emission = h5[d]['emission']

                X_data = {}
                meta_table = None
                meta_data = {}
                for k, v in data[i].items():
                    dims = len(v.shape)
                    _, spectra, meta_table = _spectra_from_image(np.transpose(np.atleast_3d(v), (1, 0, 2)),
                                                                 None, x_locs, y_locs)
                    if dims == len(axes[i]) + 1 and self.sheet in [k, "All"]:
                        # sdd-type 3D data
                        X_data[k] = spectra
                    elif dims == len(axes[i]):
                        # mapping scalars
                        meta_data[k] = spectra
                X = np.concatenate(list(X_data.values()), axis=-1)
                X_entries.append(X)

                features = np.arange(X.shape[-1])
                if len(features) == len(emission):
                    features = np.array(emission)
                features_entries.append(features)

                meta_table = meta_table.add_column(ContinuousVariable("en"), np.ones(X.shape[0]) * en, to_metas=True)
                for k, v in meta_data.items():
                    meta_table = meta_table.add_column(ContinuousVariable(k), v[:, 0], to_metas=True)
                meta_table_entries.append(meta_table)

            if not all(np.array_equal(f, features_entries[0]) for f in features_entries):
                warnings.warn("Multiple NXdata entries with incompatible shape.")
            try:
                X = np.vstack(X_entries)
            except:
                warnings.warn("Multiple NXdata entries with incompatible shape.\nLoading first entry only.")
                return features_entries[0], X_entries[0], meta_table_entries[0]
            meta_table = meta_table.concatenate(meta_table_entries)

            return features_entries[0], X, meta_table



class PreprocessorEditorRegistry:

    def __init__(self):
        self.registered = []

    def register(self, editor, priority=1000):
        self.registered.append((editor, priority))

    def sorted(self, category=None):
        if category is None:
            for editor, _ in sorted(self.registered, key=lambda x: x[1]):
                yield editor
        else:
            for editor, _ in sorted(self.registered, key=lambda x: x[1]):
                if category in editor.categories:
                    yield editor

    def categories(self):
        categories = set()
        for editor, _ in self.registered:
            categories.update(editor.categories)
        return sorted(categories)


preprocess_editors = PreprocessorEditorRegistry()



import os
import unittest
from unittest.mock import patch
import io
from base64 import b64decode

import numpy as np
from PIL import Image

try:
    import dask
    from Orange.tests.test_dasktable import temp_dasktable
except ImportError:
    dask = None

from AnyQt.QtCore import QPointF, Qt, QRectF
from AnyQt.QtTest import QSignalSpy
import Orange
from Orange.data import DiscreteVariable, Domain, Table
from Orange.widgets.tests.base import WidgetTest
from Orange.util import OrangeDeprecationWarning

from orangecontrib.spectroscopy.data import _spectra_from_image, build_spec_table
from orangecontrib.spectroscopy.io.util import VisibleImage
from orangecontrib.spectroscopy.preprocess.integrate import IntegrateFeaturePeakSimple, \
    Integrate, IntegrateFeatureSimple
from orangecontrib.spectroscopy.widgets import owhyper
from orangecontrib.spectroscopy.widgets.owhyper import \
    OWHyper
from orangecontrib.spectroscopy.preprocess import Interpolate
from orangecontrib.spectroscopy.widgets.line_geometry import in_polygon, is_left
from orangecontrib.spectroscopy.tests.util import hold_modifiers, set_png_graph_save
from orangecontrib.spectroscopy.utils import values_to_linspace, \
    index_values, location_values, index_values_nan

NAN = float("nan")


def wait_for_image(widget, timeout=5000):
    spy = QSignalSpy(widget.imageplot.image_updated)
    assert spy.wait(timeout), "Failed update image in the specified timeout"


class TestReadCoordinates(unittest.TestCase):

    def test_linspace(self):
        v = values_to_linspace(np.array([1, 2, 3]))
        np.testing.assert_equal(np.linspace(*v), [1, 2, 3])
        v = values_to_linspace(np.array([1, 2, 3, np.nan]))
        np.testing.assert_equal(np.linspace(*v), [1, 2, 3])
        v = values_to_linspace(np.array([1]))
        np.testing.assert_equal(np.linspace(*v), [1])
        v = values_to_linspace(np.array([1.001, 2, 3.002]))
        np.testing.assert_equal(np.linspace(*v), [1.001, 2.0015, 3.002])

    def test_index(self):
        a = np.array([1, 2, 3])
        v = values_to_linspace(a)
        iv = index_values(a, v)
        np.testing.assert_equal(iv, [0, 1, 2])
        a = np.array([1, 2, 3, 4])
        v = values_to_linspace(a)
        iv = index_values(a, v)
        np.testing.assert_equal(iv, [0, 1, 2, 3])
        a = np.array([1, 2, 3, 6, 5])
        v = values_to_linspace(a)
        iv = index_values(a, v)
        np.testing.assert_equal(iv, [0, 1, 2, 5, 4])

    def test_index_nan(self):
        a = np.array([1, 2, 3, np.nan])
        v = values_to_linspace(a)
        iv, nans = index_values_nan(a, v)
        np.testing.assert_equal(iv[:-1], [0, 1, 2])
        np.testing.assert_equal(nans, [0, 0, 0, 1])

    def test_location(self):
        lsc = values_to_linspace(np.array([1, 1, 1]))  # a constant
        lv = location_values([0, 1, 2], lsc)
        np.testing.assert_equal(lv, [-1, 0, 1])


class TestPolygonSelection(unittest.TestCase):

    def test_is_left(self):
        self.assertGreater(is_left(0, 0, 0, 1, -1, 0), 0)
        self.assertLess(is_left(0, 0, 0, -1, -1, 0), 0)

    def test_point(self):
        poly = [(0, 1), (1, 0), (2, 1), (3, 0), (3, 2), (0, 1)]  # non-convex

        self.assertFalse(in_polygon([0, 0], poly))
        self.assertTrue(in_polygon([1, 1.1], poly))
        self.assertTrue(in_polygon([1, 1], poly))
        self.assertTrue(in_polygon([1, 0.5], poly))
        self.assertFalse(in_polygon([2, 0], poly))
        self.assertFalse(in_polygon([0, 2], poly))

        # multiple points at once
        np.testing.assert_equal([False, True], in_polygon([[0, 0], [1, 1]], poly))

    def test_order(self):
        poly = [(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)]  # square
        self.assertTrue(in_polygon([0.5, 0.5], poly))
        self.assertTrue(in_polygon([0.5, 0.5], list(reversed(poly))))


class TestOWHyper(WidgetTest):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.iris = Orange.data.Table("iris")
        cls.whitelight = Orange.data.Table("whitelight.gsf")
        cls.whitelight_unknown = cls.whitelight.copy()
        with cls.whitelight_unknown.unlocked():
            cls.whitelight_unknown[0][0] = NAN
        # dataset with a single attribute
        cls.iris1 = cls.iris.transform(Orange.data.Domain(cls.iris.domain[:1]))
        # dataset without any attributes
        iris0 = cls.iris.transform(Orange.data.Domain([]))
        # dataset without rows
        empty = cls.iris[:0]
        # dataset with large blank regions
        irisunknown = Interpolate(np.arange(20))(cls.iris)[:20]
        # dataset without any attributes, but XY
        whitelight0 = cls.whitelight.transform(
            Orange.data.Domain([], None, metas=cls.whitelight.domain.metas))[:100]
        unknowns = cls.iris[::10].copy()
        with unknowns.unlocked():
            unknowns.X[:, :] = float("nan")
        single_pixel = cls.iris[:50]  # all image coordinates map to one spot
        unknown_pixel = single_pixel.copy()
        with unknown_pixel.unlocked():
            unknown_pixel.Y[:] = float("nan")
        cls.strange_data = [None, cls.iris1, iris0, empty, irisunknown, whitelight0,
                            unknowns, single_pixel, unknown_pixel]

    def setUp(self):
        self.widget = self.create_widget(OWHyper)  # type: OWHyper

    def test_feature_init(self):
        self.send_signal("Data", self.iris)
        self.assertEqual(self.widget.attr_value, self.iris.domain.class_var)
        attr1, attr2, attr3 = self.iris.domain.attributes[:3]
        self.assertEqual(self.widget.rgb_red_value, attr1)
        self.assertEqual(self.widget.rgb_green_value, attr2)
        self.assertEqual(self.widget.rgb_blue_value, attr3)
        self.send_signal("Data", self.iris1)
        self.assertEqual(self.widget.attr_value, attr1)
        self.assertEqual(self.widget.rgb_red_value, attr1)
        self.assertEqual(self.widget.rgb_green_value, attr1)
        self.assertEqual(self.widget.rgb_blue_value, attr1)

    def test_integral_lines(self):
        w = self.widget
        self.send_signal(OWHyper.Inputs.data, self.iris)
        icombo = w.controls.integration_method
        for i in range(icombo.count()):
            icombo.setCurrentIndex(i)
            icombo.activated.emit(i)
            wait_for_image(w)
            imethod = w.integration_methods[w.integration_method]
            if imethod == Integrate.PeakAt:
                correct = [False, False, True, False, False]
            elif imethod == Integrate.Separate:
                correct = [True, True, False, True, True]
            else:
                correct = [True, True, False, False, False]
            visible = [a.isVisible() for a in [w.line1, w.line2, w.line3, w.line4, w.line5]]
            self.assertEqual(visible, correct)

    def try_big_selection(self):
        self.widget.imageplot.select_square(QPointF(-100, -100), QPointF(100, 100))
        self.widget.imageplot.make_selection(None)

    def test_strange(self):
        for data in self.strange_data:
            self.widget = self.create_widget(OWHyper)
            self.send_signal("Data", data)
            wait_for_image(self.widget)
            self.try_big_selection()

    def test_context_not_open_invalid(self):
        self.send_signal("Data", self.iris1)
        self.assertIsNone(self.widget.imageplot.attr_x)
        self.send_signal("Data", self.iris)
        self.assertIsNotNone(self.widget.imageplot.attr_x)

    def test_no_samples(self):
        self.send_signal("Data", self.whitelight[:0])
        self.try_big_selection()

    def test_few_samples(self):
        self.send_signal("Data", self.whitelight[:1])
        self.send_signal("Data", self.whitelight[:2])
        self.send_signal("Data", self.whitelight[:3])
        self.try_big_selection()

    def test_simple(self):
        self.send_signal("Data", self.whitelight)
        self.send_signal("Data", None)
        self.try_big_selection()
        self.assertIsNone(self.get_output("Selection"), None)

    def test_unknown(self):
        self.send_signal("Data", self.whitelight[:10])
        wait_for_image(self.widget)
        levels = self.widget.imageplot.img.levels
        self.send_signal("Data", self.whitelight_unknown[:10])
        wait_for_image(self.widget)
        levelsu = self.widget.imageplot.img.levels
        np.testing.assert_equal(levelsu, levels)

    def test_select_all(self):
        self.send_signal("Data", self.whitelight)
        wait_for_image(self.widget)

        out = self.get_output("Selection")
        self.assertIsNone(out, None)

        # select all
        self.widget.imageplot.select_square(QPointF(-100, -100), QPointF(1000, 1000))
        out = self.get_output("Selection")
        self.assertEqual(len(self.whitelight), len(out))

        # test if mixing increasing and decreasing works
        self.widget.imageplot.select_square(QPointF(1000, -100), QPointF(-100, 1000))
        out = self.get_output("Selection")
        self.assertEqual(len(self.whitelight), len(out))

        # deselect
        self.widget.imageplot.select_square(QPointF(-100, -100), QPointF(-100, -100))
        out = self.get_output("Selection")
        self.assertIsNone(out, None)

        # select specific points
        self.widget.imageplot.select_square(QPointF(53.20, 30.0), QPointF(53.205, 30.02))
        out = self.get_output("Selection")
        np.testing.assert_almost_equal(out.metas,
                                       [[53.2043, 30.0185], [53.2043, 30.0085]],
                                       decimal=3)
        np.testing.assert_equal([o[out.domain["Group"]].value for o in out],
                                ["G1", "G1"])

    def test_select_polygon_as_rectangle(self):
        # rectangle and a polygon need to give the same results
        self.send_signal("Data", self.whitelight)
        wait_for_image(self.widget)
        self.widget.imageplot.select_square(QPointF(53, 30), QPointF(54, 31))
        out = self.get_output("Selection")
        self.widget.imageplot.select_polygon([QPointF(53, 30), QPointF(53, 31), QPointF(54, 31),
                                              QPointF(54, 30), QPointF(53, 30)])
        outpoly = self.get_output("Selection")
        self.assertEqual(list(out), list(outpoly))

    def test_select_click(self):
        self.send_signal("Data", self.whitelight)
        wait_for_image(self.widget)
        self.widget.imageplot.select_by_click(QPointF(53.2443, 30.6984))
        out = self.get_output("Selection")
        np.testing.assert_almost_equal(out.metas, [[53.2443, 30.6984]], decimal=3)

    def test_select_line_change_file(self):
        self.send_signal("Data", self.whitelight)
        wait_for_image(self.widget)
        # select whole image row
        self.widget.imageplot.select_line(QPointF(50, 30.6), QPointF(55, 30.6))
        out = self.get_output("Selection")
        self.assertEqual(len(out), 200)
        self.send_signal("Data", self.iris)
        wait_for_image(self.widget)
        out = self.get_output("Selection")
        self.assertIsNone(out, None)

    def test_select_click_multiple_groups(self):
        data = self.whitelight
        self.send_signal("Data", data)
        wait_for_image(self.widget)
        self.widget.imageplot.select_by_click(QPointF(53.2, 30))
        with hold_modifiers(self.widget, Qt.ShiftModifier):
            self.widget.imageplot.select_by_click(QPointF(53.4, 30))
        with hold_modifiers(self.widget, Qt.ShiftModifier):
            self.widget.imageplot.select_by_click(QPointF(53.6, 30))
        with hold_modifiers(self.widget, Qt.ControlModifier):
            self.widget.imageplot.select_by_click(QPointF(53.8, 30))
        out = self.get_output(self.widget.Outputs.annotated_data)
        self.assertEqual(len(out), 20000)  # have a data table at the output
        newvars = out.domain.variables + out.domain.metas
        oldvars = data.domain.variables + data.domain.metas
        group_at = [a for a in newvars if a not in oldvars][0]
        unselected = group_at.to_val("Unselected")
        out = out[np.asarray(np.flatnonzero(
            out.transform(Orange.data.Domain([group_at])).X != unselected))]
        self.assertEqual(len(out), 4)
        np.testing.assert_almost_equal([o["map_x"].value for o in out],
                                       [53.1993, 53.3993, 53.5993, 53.7993],
                                       decimal=3)
        np.testing.assert_equal([o[group_at].value for o in out], ["G1", "G2", "G3", "G3"])
        out = self.get_output(self.widget.Outputs.selected_data)
        np.testing.assert_equal([o[out.domain["Group"]].value for o in out],
                                ["G1", "G2", "G3", "G3"])

        # remove one element
        with hold_modifiers(self.widget, Qt.AltModifier):
            self.widget.imageplot.select_by_click(QPointF(53.2, 30))
        out = self.get_output(self.widget.Outputs.selected_data)
        np.testing.assert_equal(len(out), 3)
        np.testing.assert_equal([o[out.domain["Group"]].value for o in out],
                                ["G2", "G3", "G3"])

    def test_select_a_curve(self):
        self.send_signal("Data", self.iris)
        self.widget.curveplot.make_selection([0])

    def test_settings_curves(self):
        self.send_signal("Data", self.iris)
        self.widget.curveplot.feature_color = self.iris.domain.class_var
        self.send_signal("Data", self.whitelight)
        self.assertEqual(self.widget.curveplot.feature_color, None)
        self.send_signal("Data", self.iris)
        self.assertEqual(self.widget.curveplot.feature_color.name, "iris")

    def test_set_variable_color(self):
        data = self.iris
        ndom = Orange.data.Domain(data.domain.attributes[:-1], data.domain.class_var,
                                  metas=[data.domain.attributes[-1]])
        data = data.transform(ndom)
        self.send_signal("Data", data)
        self.widget.controls.value_type.buttons[1].click()
        with patch("orangecontrib.spectroscopy.widgets.owhyper.ImageItemNan.setLookupTable") as p:
            # a discrete variable
            self.widget.attr_value = data.domain["iris"]
            self.widget.imageplot.update_color_schema()
            self.widget.update_feature_value()
            wait_for_image(self.widget)
            np.testing.assert_equal(len(p.call_args[0][0]), 3)  # just 3 colors for 3 values
            # a continuous variable
            self.widget.attr_value = data.domain["petal width"]
            self.widget.imageplot.update_color_schema()
            self.widget.update_feature_value()
            wait_for_image(self.widget)
            np.testing.assert_equal(len(p.call_args[0][0]), 256)  # 256 for a continuous variable

    def test_color_variable_levels(self):
        class_values = ["a"], ["a", "b", "c"]
        correct_levels = [0, 0], [0, 2]
        for values, correct in zip(class_values, correct_levels):
            domain = Domain([], DiscreteVariable("c", values=values))
            data = Table.from_numpy(domain, X=[[]], Y=[[0]])
            self.send_signal("Data", data)
            self.widget.controls.value_type.buttons[1].click()
            self.widget.attr_value = data.domain.class_var
            self.widget.update_feature_value()
            wait_for_image(self.widget)
            np.testing.assert_equal(self.widget.imageplot.img.levels, correct)

    def test_single_update_view(self):
        with patch("orangecontrib.spectroscopy.widgets.owhyper.ImagePlot.update_view") as p:
            self.send_signal("Data", self.iris)
            self.assertEqual(p.call_count, 1)

    def test_correct_legend(self):
        self.send_signal("Data", self.iris)
        wait_for_image(self.widget)
        self.assertTrue(self.widget.imageplot.legend.isVisible())
        self.widget.controls.value_type.buttons[1].click()
        wait_for_image(self.widget)
        self.assertFalse(self.widget.imageplot.legend.isVisible())

    def test_migrate_selection(self):
        c = QPointF()  # some we set an attribute to
        setattr(c, "selection", [False, True, True, False])
        settings = {"context_settings": [c]}
        OWHyper.migrate_settings(settings, 2)
        self.assertEqual(settings["imageplot"]["selection_group_saved"], [(1, 1), (2, 1)])

    def test_color_no_data(self):
        self.send_signal("Data", None)
        self.widget.controls.value_type.buttons[1].click()
        self.widget.imageplot.update_color_schema()

    def test_image_too_big_error(self):
        oldval = owhyper.IMAGE_TOO_BIG
        try:
            owhyper.IMAGE_TOO_BIG = 3
            self.send_signal("Data", self.iris)
            wait_for_image(self.widget)
            self.assertTrue(self.widget.Error.image_too_big.is_shown())
        finally:
            owhyper.IMAGE_TOO_BIG = oldval

    def test_save_graph(self):
        self.send_signal("Data", self.iris)
        with set_png_graph_save() as fname:
            self.widget.save_graph()
            self.assertGreater(os.path.getsize(fname), 1000)

    def test_unknown_values_axes(self):
        data = self.iris.copy()
        with data.unlocked():
            data.Y[0] = np.nan
        self.send_signal("Data", data)
        wait_for_image(self.widget)
        self.assertTrue(self.widget.Information.not_shown.is_shown())

    def test_migrate_context_feature_color(self):
        # avoid class_vars in tests, because the setting does not allow them
        iris = self.iris.transform(
            Domain(self.iris.domain.attributes, None, self.iris.domain.class_vars))
        c = self.widget.settingsHandler.new_context(iris.domain, None, iris.domain.metas)
        c.values["curveplot"] = {"feature_color": ("iris", 1)}
        self.widget = self.create_widget(OWHyper,
                                         stored_settings={"context_settings": [c]})
        self.send_signal("Data", iris)
        self.assertIsInstance(self.widget.curveplot.feature_color, DiscreteVariable)

    def test_image_computation(self):
        spectra = [[[0, 0, 2, 0],
                    [0, 0, 1, 0]],
                   [[1, 2, 2, 0],
                    [0, 1, 1, 0]]]
        wns = [0, 1, 2, 3]
        x_locs = [0, 1]
        y_locs = [0, 1]
        data = build_spec_table(*_spectra_from_image(spectra, wns, x_locs, y_locs))

        def last_called_array(m):
            arrays = [a[0][0] for a in m.call_args_list
                      if a and a[0] and isinstance(a[0][0], np.ndarray)]
            return arrays[-1]

        wrap = self.widget.imageplot.img

        # integral from zero
        self.widget.integration_method = \
            self.widget.integration_methods.index(IntegrateFeatureSimple)
        self.send_signal("Data", data)
        with patch.object(wrap, 'setImage', wraps=wrap.setImage) as m:
            wait_for_image(self.widget)
            called = last_called_array(m)
            target = [[2, 1], [4.5, 2]]
            np.testing.assert_equal(called.squeeze(), target)

        # peak from zero
        self.widget.integration_method = \
            self.widget.integration_methods.index(IntegrateFeaturePeakSimple)
        self.widget._change_integral_type()
        with patch.object(wrap, 'setImage', wraps=wrap.setImage) as m:
            wait_for_image(self.widget)
            called = last_called_array(m)
            target = [[2, 1], [2, 1]]
            np.testing.assert_equal(called.squeeze(), target)

        # single wavenumber (feature)
        self.widget.controls.value_type.buttons[1].click()
        self.widget.attr_value = data.domain.attributes[1]
        self.widget.update_feature_value()
        with patch.object(wrap, 'setImage', wraps=wrap.setImage) as m:
            wait_for_image(self.widget)
            called = last_called_array(m)
            target = [[0, 0], [2, 1]]
            np.testing.assert_equal(called.squeeze(), target)

        # RGB
        self.widget.controls.value_type.buttons[2].click()
        self.widget.rgb_red_value = data.domain.attributes[0]
        self.widget.rgb_green_value = data.domain.attributes[1]
        self.widget.rgb_blue_value = data.domain.attributes[2]
        self.widget.update_rgb_value()
        with patch.object(wrap, 'setImage', wraps=wrap.setImage) as m:
            wait_for_image(self.widget)
            called = last_called_array(m)
            # first three wavenumbers (features) should be passed to setImage
            target = [data.X[0, :3], data.X[1, :3]], [data.X[2, :3], data.X[3, :3]]
            np.testing.assert_equal(called, target)

    def test_scatterplot_computation(self):
        spectra = [[[0, 0, 2, 0],
                    [0, 0, 1, 0]],
                   [[1, 2, 2, 0],
                    [0, 1, 1, 0]]]
        wns = [0, 1, 2, 3]
        x_locs = [0, 1]
        y_locs = [0, 1]
        data = build_spec_table(*_spectra_from_image(spectra, wns, x_locs, y_locs))

        def colors_from_brush(brushes):
            ret = []
            for b in brushes:
                c = b.color()
                ret.append((c.red(), c.blue(), c.green()))
            return ret

        wrap = self.widget.imageplot.scatterplot_item

        self.widget.imageplot.controls.draw_as_scatterplot.click()

        # integral from zero
        self.widget.integration_method = \
            self.widget.integration_methods.index(IntegrateFeatureSimple)
        self.send_signal("Data", data)
        with patch.object(wrap, 'setData', wraps=wrap.setData) as m:
            wait_for_image(self.widget)
            self.widget.imageplot.draw_scatterplot()
            call = m.call_args_list[-1]
            np.testing.assert_equal(call.kwargs['x'], [0, 1, 0, 1])
            np.testing.assert_equal(call.kwargs['y'], [0, 0, 1, 1])
            # the current state hardcoded
            np.testing.assert_equal(colors_from_brush(call.kwargs['brush']),
                                    [(0, 177, 80), (0, 124, 12), (255, 35, 241), (0, 177, 80)])

        # single wavenumber (feature)
        self.widget.controls.value_type.buttons[1].click()
        self.widget.attr_value = data.domain.attributes[1]
        self.widget.update_feature_value()
        with patch.object(wrap, 'setData', wraps=wrap.setData) as m:
            wait_for_image(self.widget)
            self.widget.imageplot.draw_scatterplot()
            call = m.call_args_list[-1]
            np.testing.assert_equal(call.kwargs['x'], [0, 1, 0, 1])
            np.testing.assert_equal(call.kwargs['y'], [0, 0, 1, 1])
            # the current state hardcoded
            np.testing.assert_equal(colors_from_brush(call.kwargs['brush']),
                                    [(0, 124, 12), (0, 124, 12), (255, 35, 241), (27, 97, 142)])

        # RGB
        self.widget.controls.value_type.buttons[2].click()
        self.widget.rgb_red_value = data.domain.attributes[0]
        self.widget.rgb_green_value = data.domain.attributes[1]
        self.widget.rgb_blue_value = data.domain.attributes[2]
        self.widget.update_rgb_value()
        with patch.object(wrap, 'setData', wraps=wrap.setData) as m:
            wait_for_image(self.widget)
            self.widget.imageplot.draw_scatterplot()
            call = m.call_args_list[-1]
            np.testing.assert_equal(call.kwargs['x'], [0, 1, 0, 1])
            np.testing.assert_equal(call.kwargs['y'], [0, 0, 1, 1])
            np.testing.assert_equal(colors_from_brush(call.kwargs['brush']),
                                    [(0, 255, 0), (0, 0, 0), (255, 255, 255), (0, 0, 128)])

    def test_migrate_visual_setttings(self):
        settings = {"curveplot":
                        {"label_title": "title",
                         "label_xaxis": "x",
                         "label_yaxis": "y"}
                    }
        OWHyper.migrate_settings(settings, 6)
        self.assertEqual(settings["visual_settings"],
                         {('Annotations', 'Title', 'Title'): 'title',
                          ('Annotations', 'x-axis title', 'Title'): 'x',
                          ('Annotations', 'y-axis title', 'Title'): 'y'})
        settings = {}
        OWHyper.migrate_settings(settings, 6)
        self.assertNotIn("visual_settings", settings)

    def test_compat_no_group(self):
        settings = {}
        OWHyper.migrate_settings(settings, 6)
        self.assertEqual(settings, {})
        self.widget = self.create_widget(OWHyper, stored_settings=settings)
        self.assertFalse(self.widget.compat_no_group)

        settings = {}
        OWHyper.migrate_settings(settings, 5)
        self.assertEqual(settings, {"compat_no_group": True})
        self.widget = self.create_widget(OWHyper, stored_settings=settings)
        self.assertTrue(self.widget.compat_no_group)



@unittest.skipUnless(dask, "installed Orange does not support dask")
class TestOWHyperWithDask(TestOWHyper):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.iris = temp_dasktable("iris")
        cls.whitelight = temp_dasktable("whitelight.gsf")
        cls.whitelight_unknown = temp_dasktable(cls.whitelight_unknown)
        cls.iris1 = temp_dasktable(cls.iris1)
        cls.strange_data = [temp_dasktable(d) if d is not None else None
                            for d in cls.strange_data]


class _VisibleImageStream(VisibleImage):
    """ Do not use this class in practice because too many things
    will get copied when transforming tables."""

    def __init__(self, name, pos_x, pos_y, size_x, size_y, stream):
        super().__init__(name, pos_x, pos_y, size_x, size_y)
        self.stream = stream

    @property
    def image(self):
        return Image.open(self.stream)


class TestVisibleImage(WidgetTest):

    @classmethod
    def mock_visible_image_data_oldformat(cls):
        red_img = io.BytesIO(b64decode("iVBORw0KGgoAAAANSUhEUgAAAA"
                                       "oAAAAKCAYAAACNMs+9AAAAFUlE"
                                       "QVR42mP8z8AARIQB46hC+ioEAG"
                                       "X8E/cKr6qsAAAAAElFTkSuQmCC"))
        black_img = io.BytesIO(b64decode("iVBORw0KGgoAAAANSUhEUgAAA"
                                         "AoAAAAKCAQAAAAnOwc2AAAAEU"
                                         "lEQVR42mNk+M+AARiHsiAAcCI"
                                         "KAYwFoQ8AAAAASUVORK5CYII="))

        return [
            {
                "name": "Image 01",
                "image_ref": red_img,
                "pos_x": 100,
                "pos_y": 100,
                "pixel_size_x": 1.7,
                "pixel_size_y": 2.3
            },
            {
                "name": "Image 02",
                "image_ref": black_img,
                "pos_x": 0.5,
                "pos_y": 0.5,
                "pixel_size_x": 1,
                "pixel_size_y": 0.3
            },
            {
                "name": "Image 03",
                "image_ref": red_img,
                "pos_x": 100,
                "pos_y": 100,
                "img_size_x": 17.0,
                "img_size_y": 23.0
            },
        ]

    @classmethod
    def mock_visible_image_data(cls):
        red_img = io.BytesIO(b64decode("iVBORw0KGgoAAAANSUhEUgAAAA"
                                       "oAAAAKCAYAAACNMs+9AAAAFUlE"
                                       "QVR42mP8z8AARIQB46hC+ioEAG"
                                       "X8E/cKr6qsAAAAAElFTkSuQmCC"))
        black_img = io.BytesIO(b64decode("iVBORw0KGgoAAAANSUhEUgAAA"
                                         "AoAAAAKCAQAAAAnOwc2AAAAEU"
                                         "lEQVR42mNk+M+AARiHsiAAcCI"
                                         "KAYwFoQ8AAAAASUVORK5CYII="))

        return [
            _VisibleImageStream("Image 01", 100, 100, 17., 23., red_img),
            _VisibleImageStream("Image 02", 0.5, 0.5, 10, 3, black_img),
            _VisibleImageStream("Image 03", 100, 100, 17., 23., red_img),
        ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.data_with_visible_images = Orange.data.Table(
            "agilent/4_noimage_agg256.dat"
        )
        cls.data_with_visible_images.attributes["visible_images"] = \
            cls.mock_visible_image_data()

    def setUp(self):
        self.widget = self.create_widget(OWHyper)  # type: OWHyper

    def assert_same_visible_image(self, img_info, vis_img, mock_rect):
        img = img_info.image.convert('RGBA')
        img = np.array(img)[::-1]
        rect = QRectF(img_info.pos_x, img_info.pos_y,
                      img_info.size_x, img_info.size_y)
        self.assertTrue((vis_img.image == img).all())
        mock_rect.assert_called_with(rect)

    def test_no_visible_image(self):
        data = Orange.data.Table("agilent/4_noimage_agg256.dat")
        self.send_signal("Data", data)
        wait_for_image(self.widget)

        self.assertFalse(self.widget.visbox.isEnabled())

    def test_controls_enabled_when_visible_image(self):
        w = self.widget
        self.send_signal("Data", self.data_with_visible_images)
        wait_for_image(w)

        self.assertTrue(w.visbox.isEnabled())

    def test_controls_enabled_by_show_chkbox(self):
        w = self.widget
        self.send_signal("Data", self.data_with_visible_images)
        wait_for_image(w)

        self.assertTrue(w.controls.show_visible_image.isEnabled())
        self.assertFalse(w.show_visible_image)
        controls = [w.controls.visible_image,
                    w.controls.visible_image_composition,
                    w.controls.visible_image_opacity]
        for control in controls:
            self.assertFalse(control.isEnabled())

        w.controls.show_visible_image.setChecked(True)
        for control in controls:
            self.assertTrue(control.isEnabled())

    def test_first_visible_image_selected_in_combobox_by_default(self):
        w = self.widget
        vis_img = w.imageplot.vis_img
        with patch.object(vis_img, 'setRect', wraps=vis_img.setRect) as mock_rect:
            data = self.data_with_visible_images
            self.send_signal("Data", data)
            wait_for_image(w)

            w.controls.show_visible_image.setChecked(True)
            self.assertEqual(len(w.visible_image_model),
                             len(data.attributes["visible_images"]))
            self.assertEqual(w.visible_image, data.attributes["visible_images"][0])
            self.assertEqual(w.controls.visible_image.currentIndex(), 0)
            self.assertEqual(w.controls.visible_image.currentText(), "Image 01")

            self.assert_same_visible_image(data.attributes["visible_images"][0],
                                           w.imageplot.vis_img,
                                           mock_rect)

    def test_visible_image_displayed(self):
        w = self.widget
        data = self.data_with_visible_images
        self.send_signal("Data", data)
        wait_for_image(w)

        self.assertNotIn(w.imageplot.vis_img, w.imageplot.plot.items)

        w.controls.show_visible_image.setChecked(True)
        self.assertIn(w.imageplot.vis_img, w.imageplot.plot.items)
        w.controls.show_visible_image.setChecked(False)
        self.assertNotIn(w.imageplot.vis_img, w.imageplot.plot.items)

    def test_hide_visible_image_after_no_image_loaded(self):
        w = self.widget
        data = self.data_with_visible_images
        self.send_signal("Data", data)
        wait_for_image(w)

        w.controls.show_visible_image.setChecked(True)
        data = Orange.data.Table("agilent/4_noimage_agg256.dat")
        self.send_signal("Data", data)
        wait_for_image(w)

        self.assertFalse(w.visbox.isEnabled())
        self.assertFalse(w.show_visible_image)
        self.assertNotIn(w.imageplot.vis_img, w.imageplot.plot.items)

    def test_select_another_visible_image(self):
        w = self.widget
        data = self.data_with_visible_images
        self.send_signal("Data", data)
        wait_for_image(w)

        w.controls.show_visible_image.setChecked(True)
        vis_img = w.imageplot.vis_img
        with patch.object(vis_img, 'setRect', wraps=vis_img.setRect) as mock_rect:
            w.controls.visible_image.setCurrentIndex(1)
            # since activated signal emitted only by visual interaction
            # we need to trigger it by hand here.
            w.controls.visible_image.activated.emit(1)

            self.assert_same_visible_image(data.attributes["visible_images"][1],
                                           w.imageplot.vis_img,
                                           mock_rect)

    def test_visible_image_opacity(self):
        w = self.widget
        data = self.data_with_visible_images
        self.send_signal("Data", data)
        wait_for_image(w)

        with patch.object(w.imageplot.vis_img, 'setOpacity') as m:
            w.controls.visible_image_opacity.setValue(20)
            self.assertEqual(w.visible_image_opacity, 20)
            m.assert_called_once_with(w.visible_image_opacity / 255)

    def test_visible_image_composition_mode(self):
        w = self.widget
        self.assertEqual(w.controls.visible_image_composition.currentText(), 'Normal')

        data = self.data_with_visible_images
        self.send_signal("Data", data)
        wait_for_image(w)

        for i in range(len(w.visual_image_composition_modes)):
            with patch.object(w.imageplot.vis_img, 'setCompositionMode') as m:
                w.controls.visible_image_composition.setCurrentIndex(i)
                # since activated signal emitted only by visual interaction
                # we need to trigger it by hand here
                w.controls.visible_image_composition.activated.emit(i)
                name = w.controls.visible_image_composition.currentText()
                mode = w.visual_image_composition_modes[name]
                m.assert_called_once_with(mode)

    def test_visible_image_img_size(self):
        w = self.widget
        data = self.data_with_visible_images
        self.send_signal("Data", data)
        wait_for_image(w)

        w.controls.show_visible_image.setChecked(True)
        vis_img = w.imageplot.vis_img
        with patch.object(vis_img, 'setRect', wraps=vis_img.setRect) as mock_rect:
            w.controls.visible_image.setCurrentIndex(2)
            # since activated signal emitted only by visual interaction
            # we need to trigger it by hand here.
            w.controls.visible_image.activated.emit(2)

            self.assert_same_visible_image(data.attributes["visible_images"][0],
                                           w.imageplot.vis_img,
                                           mock_rect)

    def test_oldformat(self):
        data = Orange.data.Table(
            "agilent/4_noimage_agg256.dat"
        )
        data.attributes["visible_images"] = \
            self.mock_visible_image_data_oldformat()
        w = self.widget

        with self.assertWarns(OrangeDeprecationWarning):
            self.send_signal("Data", data)

        wait_for_image(w)

        self.assertNotIn(w.imageplot.vis_img, w.imageplot.plot.items)

        w.controls.show_visible_image.setChecked(True)
        self.assertIn(w.imageplot.vis_img, w.imageplot.plot.items)
        w.controls.show_visible_image.setChecked(False)
        self.assertNotIn(w.imageplot.vis_img, w.imageplot.plot.items)


class TestVectorPlot(WidgetTest):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.whitelight = Orange.data.Table("whitelight.gsf")
        cls.iris = Orange.data.Table("iris")

    def setUp(self):
        super().setUp()
        self.widget = self.create_widget(OWHyper)  # type: OWHyper

    def test_enable_disable(self):
        w = self.widget
        for data in [None, self.whitelight, self.iris]:
            self.send_signal(w.Inputs.data, data)
            self.assertFalse(w.imageplot.show_vector_plot)
            self.assertFalse(w.controls.imageplot.vector_magnitude.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_angle.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_color_index.isEnabled())
            self.assertFalse(w.controls.imageplot.vcol_byval_index.isEnabled())
            self.assertFalse(w.controls.imageplot.vcol_byval_feat.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_scale.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_width.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_opacity.isEnabled())
            self.assertFalse(w.controls.imageplot.v_bin.isEnabled())

            w.controls.imageplot.show_vector_plot.click()
            self.assertTrue(w.imageplot.show_vector_plot)
            self.assertTrue(w.controls.imageplot.vector_magnitude.isEnabled())
            self.assertTrue(w.controls.imageplot.vector_angle.isEnabled())
            self.assertTrue(w.controls.imageplot.vector_color_index.isEnabled())
            self.assertTrue(w.controls.imageplot.vector_scale.isEnabled())
            self.assertTrue(w.controls.imageplot.vector_width.isEnabled())
            self.assertTrue(w.controls.imageplot.vector_opacity.isEnabled())
            self.assertTrue(w.controls.imageplot.v_bin.isEnabled())

            w.imageplot.vector_color_index = 8
            w.imageplot._update_vector()
            self.assertTrue(w.controls.imageplot.vcol_byval_index.isEnabled())
            self.assertTrue(w.controls.imageplot.vcol_byval_feat.isEnabled())

            w.imageplot.vector_color_index = 3
            w.imageplot._update_vector()
            self.assertFalse(w.controls.imageplot.vcol_byval_index.isEnabled())
            self.assertFalse(w.controls.imageplot.vcol_byval_feat.isEnabled())

            w.imageplot.vector_color_index = 8
            w.imageplot._update_vector()
            self.assertTrue(w.controls.imageplot.vcol_byval_index.isEnabled())
            self.assertTrue(w.controls.imageplot.vcol_byval_feat.isEnabled())

            w.controls.imageplot.show_vector_plot.click()
            self.assertFalse(w.controls.imageplot.vector_magnitude.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_angle.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_color_index.isEnabled())
            self.assertFalse(w.controls.imageplot.vcol_byval_index.isEnabled())
            self.assertFalse(w.controls.imageplot.vcol_byval_feat.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_scale.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_width.isEnabled())
            self.assertFalse(w.controls.imageplot.vector_opacity.isEnabled())
            self.assertFalse(w.controls.imageplot.v_bin.isEnabled())

    def test_legend(self):
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.widget.controls.imageplot.show_vector_plot.setChecked(True)
        self.widget.imageplot.enable_vector()
        self.widget.imageplot.vector_color_index = 8
        self.widget.imageplot._update_vector()
        self.assertFalse(self.widget.imageplot.vect_legend.isVisible())
        self.widget.imageplot.vcol_byval_feat = self.iris.domain.attributes[0]
        self.widget.imageplot._update_cbyval()
        self.assertTrue(self.widget.imageplot.vect_legend.isVisible())
        self.widget.imageplot.vcol_byval_feat = None
        self.widget.imageplot._update_cbyval()
        self.assertFalse(self.widget.imageplot.vect_legend.isVisible())
        self.widget.controls.imageplot.show_vector_plot.setChecked(False)
        self.widget.imageplot.enable_vector()
    
    def test_vect_color(self):
        feat = self.iris.get_column(self.iris.domain.attributes[0])
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.widget.controls.imageplot.show_vector_plot.setChecked(True)
        self.widget.imageplot.enable_vector()
        for i in range(8):
            self.widget.imageplot.vector_color_index = i
            self.widget.imageplot._update_vector()
            self.assertEqual(len(self.widget.imageplot.get_vector_color(feat)), 4)
        self.widget.imageplot.vector_color_index = 8
        self.widget.imageplot._update_vector()
        self.assertEqual(self.widget.imageplot.get_vector_color(feat)[0].shape, (feat.shape[0], 4))
        self.widget.controls.imageplot.show_vector_plot.setChecked(False)
        self.widget.imageplot.enable_vector()

    def test_vect_bin(self):
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.widget.controls.imageplot.show_vector_plot.setChecked(True)
        self.widget.imageplot.enable_vector()
        self.widget.imageplot.vector_angle = self.iris.domain.attributes[0]
        self.widget.imageplot.vector_magnitude = self.iris.domain.attributes[0]
        self.widget.imageplot._update_vector_params()

        self.widget.imageplot.v_bin = 0
        self.widget.imageplot._update_binsize()
        self.widget.imageplot.update_view()
        wait_for_image(self.widget)
        print(self.widget.imageplot.vector_plot.params[0].shape)
        self.assertEqual(self.widget.imageplot.vector_plot.params[0].shape[0],
                         self.iris.X.shape[0]*2)
        self.assertEqual(self.widget.imageplot.vector_plot.params[1].shape[0],
                         self.iris.X.shape[0]*2)

        self.widget.imageplot.v_bin = 1
        self.widget.imageplot._update_binsize()
        self.widget.imageplot.update_view()
        wait_for_image(self.widget)
        self.assertEqual(self.widget.imageplot.vector_plot.params[0].shape[0], 2)
        self.assertEqual(self.widget.imageplot.vector_plot.params[1].shape[0], 2)

        self.widget.imageplot.v_bin = 2
        self.widget.imageplot._update_binsize()
        self.widget.imageplot.update_view()
        wait_for_image(self.widget)
        self.assertEqual(self.widget.imageplot.vector_plot.params[0].shape[0], 2)
        self.assertEqual(self.widget.imageplot.vector_plot.params[1].shape[0], 2)
        
        self.widget.imageplot.v_bin = 3
        self.widget.imageplot._update_binsize()
        self.widget.imageplot.update_view()
        wait_for_image(self.widget)
        self.assertTrue(self.widget.Warning.bin_size_error.is_shown())
        self.assertEqual(self.widget.imageplot.vector_plot.params[0].shape[0], 2)
        self.assertEqual(self.widget.imageplot.vector_plot.params[1].shape[0], 2)



import numpy as np
import sklearn.cross_decomposition as skl_pls

from Orange.data import Table, Domain, Variable, \
    ContinuousVariable, StringVariable
from Orange.data.util import get_unique_names, SharedComputeValue
from Orange.preprocess.score import LearnerScorer
from Orange.regression import SklLearner, SklModel


class _FeatureScorerMixin(LearnerScorer):
    feature_type = Variable
    class_type = ContinuousVariable

    def score(self, data):
        model = self(data)
        return np.abs(model.coefficients), model.domain.attributes


class _PLSCommonTransform:

    def __init__(self, pls_model):
        self.pls_model = pls_model

    def _transform_with_numpy_output(self, X, Y):
        pls = self.pls_model.skl_model
        """
        # the next command does the following
        x_center = X - pls._x_mean
        y_center = Y - pls._y_mean
        t = x_center @ pls.x_rotations_
        u = y_center @ pls.y_rotations_
        """
        t, u = pls.transform(X, Y)
        return np.hstack((t, u))

    def __call__(self, data):
        if data.domain != self.pls_model.domain:
            data = data.transform(self.pls_model.domain)
        if len(data.Y.shape) == 1:
            Y = data.Y.reshape(-1, 1)
        else:
            Y = data.Y
        return self._transform_with_numpy_output(data.X, Y)


class PLSProjector(SharedComputeValue):
    def __init__(self, transform, feature):
        super().__init__(transform)
        self.feature = feature

    def compute(self, _, space):
        return space[:, self.feature]


class PLSModel(SklModel):
    var_prefix_X = "PLS T"
    var_prefix_Y = "PLS U"

    @property
    def coefficients(self):
        coef = self.skl_model.coef_
        return coef

    def predict(self, X):
        vals = self.skl_model.predict(X)
        if len(self.domain.class_vars) == 1:
            vals = vals.ravel()
        return vals

    def __str__(self):
        return 'PLSModel {}'.format(self.skl_model)

    def _get_var_names(self, n, prefix):
        names = [f"{prefix}{postfix}" for postfix in range(1, n + 1)]
        return get_unique_names([var.name for var in self.domain.metas], names)

    def project(self, data):
        if not isinstance(data, Table):
            raise RuntimeError("PLSModel can only project tables")

        transformer = _PLSCommonTransform(self)

        def trvar(i, name):
            return ContinuousVariable(name, compute_value=PLSProjector(transformer, i))

        n_components = self.skl_model.x_loadings_.shape[1]

        var_names_X = self._get_var_names(n_components, self.var_prefix_X)
        var_names_Y = self._get_var_names(n_components, self.var_prefix_Y)

        domain = Domain(
            [trvar(i, var_names_X[i]) for i in range(n_components)],
            data.domain.class_vars,
            list(data.domain.metas) +
            [trvar(n_components + i, var_names_Y[i]) for i in range(n_components)]
        )

        return data.transform(domain)

    def components(self):
        orig_domain = self.domain
        names = [a.name for a in orig_domain.attributes + orig_domain.class_vars]
        meta_name = get_unique_names(names, 'components')

        n_components = self.skl_model.x_loadings_.shape[1]

        meta_vars = [StringVariable(name=meta_name)]
        metas = np.array(
            [[f"Component {i + 1}" for i in range(n_components)]], dtype=object
        ).T
        dom = Domain(
            [ContinuousVariable(a.name) for a in orig_domain.attributes],
            [ContinuousVariable(a.name) for a in orig_domain.class_vars],
            metas=meta_vars)
        components = Table(dom,
                           self.skl_model.x_loadings_.T,
                           Y=self.skl_model.y_loadings_.T,
                           metas=metas)
        components.name = 'components'
        return components

    def coefficients_table(self):
        coeffs = self.coefficients.T
        domain = Domain(
            [ContinuousVariable(f"coef {i}") for i in range(coeffs.shape[1])],
            metas=[StringVariable("name")]
        )
        waves = [[attr.name] for attr in self.domain.attributes]
        coef_table = Table.from_numpy(domain, X=coeffs, metas=waves)
        coef_table.name = "coefficients"
        return coef_table


class PLSRegressionLearner(SklLearner, _FeatureScorerMixin):
    __wraps__ = skl_pls.PLSRegression
    __returns__ = PLSModel
    supports_multiclass = True
    preprocessors = SklLearner.preprocessors

    def fit(self, X, Y, W=None):
        params = self.params.copy()
        params["n_components"] = min(X.shape[1] - 1,
                                     X.shape[0] - 1,
                                     params["n_components"])
        clf = self.__wraps__(**params)
        return self.__returns__(clf.fit(X, Y))

    def __init__(self, n_components=2, scale=True,
                 max_iter=500, preprocessors=None):
        super().__init__(preprocessors=preprocessors)
        self.params = vars()

    def incompatibility_reason(self, domain):
        reason = None
        if not domain.class_vars:
            reason = "Numeric targets expected."
        else:
            for cv in domain.class_vars:
                if not cv.is_continuous:
                    reason = "Only numeric target variables expected."
        return reason

if __name__ == '__main__':
    import Orange

    data = Orange.data.Table('housing')
    learners = [PLSRegressionLearner(n_components=2, max_iter=100)]
    res = Orange.evaluation.CrossValidation()(data, learners)
    for l, ca in zip(learners, Orange.evaluation.RMSE(res)):
        print("learner: {}\nRMSE: {}\n".format(l, ca))



import os

import numpy as np

from Orange.data import Table, Domain, FileFormat

from orangecontrib.spectroscopy.io.util import SpectralFileFormat, _spectra_from_image_2d
from orangecontrib.spectroscopy.utils.specio.specio import BlockReader, PerkinElmer

# This code is partially based on software developed in the Diamond synchrotron.


class PerkinElmerReader(FileFormat, SpectralFileFormat):
    EXTENSIONS = (".sp", ".fsm",)
    DESCRIPTION = "Perkin Elmer"

    def read_sp(self):
        with open(self.filename, "rb") as f:
            data = f.read()

        reader = BlockReader(data)

        info = {
            "signature": reader.read(4, format="utf-8"),
            "description": reader.read(40, format="utf-8"),
        }

        decoders = {
            25739: PerkinElmer.decode25739,
            35698: PerkinElmer.decode35698,
            35699: PerkinElmer.decode35699,
            35700: PerkinElmer.decode35700,
            35701: PerkinElmer.decode35701,
            35708: PerkinElmer.decode35708,
        }

        stops = []
        spectrum = []

        block_id, block_size = reader.read(6,
                                           format="<Hi",
                                           expect_tuple=True)

        stops.append(reader.start + block_size)

        while block_id != 122 and not reader.atEnd(2):
            next_block_id = reader.peek(2)

            if next_block_id[1] == 117:
                reader.start = stops[-1]
                stops = stops[:-1]

                while reader.start >= stops[-1]:
                    stops = stops[:-1]

            else:
                block_id, block_size = reader.read(6,
                                                   format="<Hi",
                                                   expect_tuple=True)

                stops.append(reader.start + block_size)

        info.update(PerkinElmer.decode5104(reader.read(block_size)))

        reader.start = stops[1]

        while not reader.atEnd():
            block_id, block_size = reader.read(6,
                                               format="<Hi",
                                               expect_tuple=True)

            if block_id in decoders:
                decoded = decoders[block_id](reader.peek(block_size))

                if isinstance(decoded, dict):
                    info.update(decoded)

                else:
                    spectrum = decoded

            reader.start += block_size

        wavenumbers = np.linspace(info['min_wavelength'],
                                  info['max_wavelength'],
                                  info['n_points'])

        datavals = np.array(spectrum)[None, ...]

        domain = Domain([], None)
        meta_data = Table.from_numpy(domain,
                                     X=np.zeros((len(datavals), 0)))

        meta_data.attributes = info

        return wavenumbers, datavals, meta_data

    def read_fsm(self):
        with open(self.filename, "rb") as f:
            data = f.read()

        reader = BlockReader(data)

        meta = {
            "signature": reader.read(4, format="utf-8"),
            "description": reader.read(40, format="utf-8"),
        }

        decoders = {
            5100: PerkinElmer.decode5100,
            5104: PerkinElmer.decode5104,
            5105: PerkinElmer.decode5105,
        }

        spectrum = []

        while not reader.atEnd(6):
            block_id, block_size = reader.read(6,
                                               format="<Hi",
                                               expect_tuple=True)

            decoded = decoders[block_id](reader.read(block_size))

            if isinstance(decoded, dict):
                meta.update(decoded)
            else:
                spectrum.append(decoded)

        wavenumbers = np.arange(meta['z_start'],
                                meta['z_end'] + meta['z_delta'],
                                meta['z_delta'])

        datavals = np.squeeze(spectrum)

        domain = Domain([], None)
        meta_data = Table.from_numpy(domain,
                                     X=np.zeros((len(datavals), 0)))

        meta_data.attributes = meta

        return datavals, wavenumbers, meta_data

    def read_spectra(self):
        if os.path.splitext(self.filename)[1].lower() == ".sp":
            return self.read_sp()

        elif os.path.splitext(self.filename)[1].lower() == ".fsm":
            intensities, wn, m = self.read_fsm()

            atts = m.attributes
            # TODO position calculation should be done by a helper function that accepts units as well
            xpositions = atts['x_init'] + atts['x_delta'] * np.arange(atts['n_x'])
            ypositions = atts['y_init'] + atts['y_delta'] * np.arange(atts['n_y'])

            # TODO this is not nice - we need new helpers to build the coordinates
            y_loc = np.repeat(np.arange(atts['n_y']), atts['n_x'])
            x_loc = np.tile(np.arange(atts['n_x']), atts['n_y'])

            features, final_data, locs_table = _spectra_from_image_2d(intensities, wn,
                                                                      xpositions[x_loc],
                                                                      ypositions[y_loc])

            return features, final_data, locs_table

        else:
            raise IOError("File can't be read: unsupported file type.")



import unittest

import numpy as np

try:
    import dask
    from Orange.tests.test_dasktable import temp_dasktable
except ImportError:
    dask = None

import Orange

from orangecontrib.spectroscopy.preprocess import Interpolate, \
    interp1d_with_unknowns_numpy, interp1d_with_unknowns_scipy, \
    interp1d_wo_unknowns_scipy, InterpolateToDomain, NotAllContinuousException, \
    nan_extend_edges_and_interpolate
from orangecontrib.spectroscopy.data import getx
from orangecontrib.spectroscopy.tests.util import spectra_table
from orangecontrib.spectroscopy.tests.test_preprocess import TestCommonIndpSamplesMixin, \
    SMALL_COLLAGEN


class TestInterpolate(unittest.TestCase, TestCommonIndpSamplesMixin):

    preprocessors = [Interpolate(np.linspace(1000, 1700, 100))]
    data = SMALL_COLLAGEN

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.iris = Orange.data.Table("iris")[:5].copy()
        cls.collagen = Orange.data.Table("collagen.csv")[:5]
        cls.titanic = Orange.data.Table("titanic")
        ys = np.arange(16, dtype=float).reshape(4, 4)
        np.fill_diagonal(ys, np.nan)
        cls.range16 = spectra_table([0, 1, 2, 3], X=ys)

    def test_nofloatname(self):
        data = self.iris
        interpolated = Interpolate([0.5])(data)
        self.assertIsInstance(interpolated, type(data))
        av1 = interpolated.X.ravel()
        av2 = data.X[:, :2].mean(axis=1)
        np.testing.assert_allclose(av1, av2)

    def test_floatname(self):
        data = self.collagen
        f1, f2 = 20, 21
        c1, c2 = float(data.domain.attributes[f1].name), \
                 float(data.domain.attributes[f2].name)
        avg = (c1 + c2)/2
        interpolated = Interpolate([avg])(data)
        self.assertIsInstance(interpolated, type(data))
        av1 = interpolated.X.ravel()
        av2 = data.X[:, [20,21]].mean(axis=1)
        np.testing.assert_allclose(av1, av2)

    def test_domain_conversion(self):
        """Test whether a domain can be used for conversion."""
        data = self.iris
        interpolated = Interpolate([0.5, 1.5])(data)
        nt = data.transform(interpolated.domain)
        self.assertEqual(interpolated.domain, nt.domain)
        np.testing.assert_equal(np.asarray(interpolated.X), np.asarray(nt.X))
        np.testing.assert_equal(np.asarray(interpolated.Y), np.asarray(nt.Y))

    def test_same(self):
        """Interpolate values are original values."""
        data = self.iris
        interpolated = Interpolate(range(len(data.domain.attributes)))(data)
        np.testing.assert_allclose(interpolated.X, data.X)

    def test_permute(self):
        rs = np.random.RandomState(0)
        data = self.iris
        oldX = data.X
        #permute data
        p = rs.permutation(range(len(data.domain.attributes)))
        nattr = [Orange.data.ContinuousVariable(str(p[i]))
                 for i, a in enumerate(data.domain.attributes)]
        data = Orange.data.Table.from_numpy(Orange.data.Domain(nattr),
                                            X=data.X[:, p])
        interpolated = Interpolate(range(len(data.domain.attributes)))(data)
        np.testing.assert_allclose(interpolated.X, oldX)
        #also permute output
        p1 = rs.permutation(range(len(data.domain.attributes)))
        interpolated = Interpolate(p1)(data)
        np.testing.assert_allclose(interpolated.X, oldX[:, p1])

    def test_out_of_band(self):
        data = self.iris
        interpolated = Interpolate(range(-1, len(data.domain.attributes)+1))(data)
        np.testing.assert_allclose(interpolated.X[:, 1:5], data.X)
        np.testing.assert_equal(np.asarray(interpolated.X[:, [0, -1]]), np.nan)

    def test_unknown_middle(self):
        data = self.iris.copy()
        # whole column in the middle should be interpolated
        with data.unlocked():
            data.X[:, 1] = np.nan
        interpolated = Interpolate(getx(data))(data)
        self.assertFalse(np.any(np.isnan(interpolated.X)))

    def test_unknown_elsewhere(self):
        data = self.iris.copy()
        with data.unlocked():
            data.X[0, 1] = np.nan
            data.X[1, 1] = np.nan
            data.X[1, 2] = np.nan
        im = Interpolate(getx(data))
        interpolated = im(data)
        self.assertAlmostEqual(interpolated.X[0, 1], 3.25)
        self.assertAlmostEqual(interpolated.X[1, 1], 3.333333333333334)
        self.assertAlmostEqual(interpolated.X[1, 2], 1.766666666666667)
        self.assertFalse(np.any(np.isnan(interpolated.X)))

    def test_unknown_elsewhere_different(self):
        data = self.iris.copy()
        with data.unlocked():
            data.X[0, 1] = np.nan
            data.X[1, 1] = np.nan
            data.X[1, 2] = np.nan
        im = Interpolate(getx(data))
        im.interpfn = interp1d_with_unknowns_numpy
        interpolated = im(data)
        self.assertIsInstance(interpolated.X, type(data.X))
        self.assertAlmostEqual(interpolated.X[0, 1], 3.25)
        self.assertAlmostEqual(interpolated.X[1, 1], 3.333333333333334)
        self.assertAlmostEqual(interpolated.X[1, 2], 1.766666666666667)
        self.assertFalse(np.any(np.isnan(interpolated.X)))
        im.interpfn = interp1d_with_unknowns_scipy
        interpolated = im(data)
        self.assertIsInstance(interpolated.X, type(data.X))
        self.assertAlmostEqual(interpolated.X[0, 1], 3.25)
        self.assertAlmostEqual(interpolated.X[1, 1], 3.333333333333334)
        self.assertAlmostEqual(interpolated.X[1, 2], 1.766666666666667)
        self.assertFalse(np.any(np.isnan(interpolated.X)))
        save_X = interpolated.X
        im.interpfn = interp1d_wo_unknowns_scipy
        interpolated = im(data)
        self.assertIsInstance(interpolated.X, type(data.X))
        self.assertTrue(np.any(np.isnan(interpolated.X)))
        # parts without unknown should be the same
        np.testing.assert_allclose(data.X[2:], save_X[2:])

    def test_nan_extend_edges_and_interpolate(self):
        data = self.iris.copy()
        with data.unlocked():
            data.X[0, :] = np.nan
            data.X[1, 1] = np.nan
            data.X[2, 0] = np.nan
            data.X[3, -1] = np.nan
        xs = getx(data)
        interp, unknowns = nan_extend_edges_and_interpolate(xs, data.X)
        self.assertIsInstance(interp, type(data.X))
        nan = float("nan")
        res = np.array([[nan, nan, nan, nan],
                        [4.9, 3.15, 1.4, 0.2],
                        [3.2, 3.2, 1.3, 0.2],
                        [4.6, 3.1, 1.5, 1.5],
                        [5., 3.6, 1.4, 0.2]])
        resu = np.array([[True, True, True, True],
                         [False, True, False, False],
                         [True, False, False, False],
                         [False, False, False, True],
                         [False, False, False, False]])
        np.testing.assert_allclose(interp, res)
        np.testing.assert_allclose(unknowns, resu)

    def test_nan_extend_edges_and_interpolate_mixed(self):
        data = self.range16

        xs = getx(data)
        ys = data.X
        v, n = nan_extend_edges_and_interpolate(xs, ys)
        exp = np.arange(16, dtype=float).reshape(4, 4)
        exp[0, 0] = 1
        exp[3, 3] = 14
        np.testing.assert_equal(v, exp)

        mix = np.array([0, 2, 1, 3])
        xsm = xs[mix]
        ysm = ys[:, mix]
        v, n = nan_extend_edges_and_interpolate(xsm, ysm)
        np.testing.assert_equal(v[:, mix], exp)

    def test_eq(self):
        data = Orange.data.Table("iris")
        i1 = Interpolate([0, 1])(data)
        i2 = Interpolate([0, 1])(data)
        self.assertEqual(i1.domain, i2.domain)
        i3 = Interpolate([0, 1.1])(data)
        self.assertNotEqual(i1.domain[0], i3.domain[0])
        i4 = Interpolate([0, 1], kind="quadratic")(data)
        self.assertNotEqual(i1.domain[0], i4.domain[0])

        # different domain
        titanic = Orange.data.Table("titanic")
        it1 = Interpolate([0, 1])(titanic)
        self.assertNotEqual(i1.domain[0], it1.domain[0])


class TestInterpolateToDomain(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.iris = Orange.data.Table("iris")
        cls.housing = Orange.data.Table("housing")
        cls.titanic = Orange.data.Table("titanic")

    def test_same_domain(self):
        iris = self.iris
        housing = self.housing
        iiris = InterpolateToDomain(target=housing)(iris)
        self.assertNotEqual(housing.domain.attributes, iris.domain.attributes)
        # needs to have the same attributes
        self.assertEqual(housing.domain.attributes, iiris.domain.attributes)
        # first 4 values are defined, the rest are not
        np.testing.assert_equal(np.isnan(iiris.X[0])[:4], False)
        np.testing.assert_equal(np.isnan(iiris.X[0])[4:], True)

    def test_not_all_continuous(self):
        titanic = self.titanic
        iris = self.iris
        InterpolateToDomain(target=iris)
        with self.assertRaises(NotAllContinuousException):
            InterpolateToDomain(target=titanic)


@unittest.skipUnless(dask, "installed Orange does not support dask")
class TestInterpolateDask(TestInterpolate):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.iris = temp_dasktable(cls.iris)
        cls.collagen = temp_dasktable(cls.collagen)
        cls.range16 = temp_dasktable(cls.range16)


@unittest.skipUnless(dask, "installed Orange does not support dask")
class TestInterpolateToDomainDask(TestInterpolateToDomain):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.iris = temp_dasktable(cls.iris)
        cls.housing = temp_dasktable(cls.housing)
        cls.titanic = temp_dasktable(cls.titanic)



import unittest
from unittest.mock import patch
import numpy as np

import Orange
from Orange.widgets.tests.base import WidgetTest
from orangecontrib.spectroscopy.widgets.owoverlay import OWOverlay
from orangecontrib.spectroscopy.tests.test_owhyper import (
    wait_for_image,
    TestVisibleImage,
)

NAN = float("nan")


class TestOWOverlay(WidgetTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # cls.iris = Orange.data.Table("iris")
        cls.whitelight = Orange.data.Table("whitelight.gsf")
        cls.whitelight_unknown = cls.whitelight.copy()
        with cls.whitelight_unknown.unlocked():
            cls.whitelight_unknown[0][0] = NAN

    def setUp(self):
        self.widget = self.create_widget(OWOverlay)  # type: OWOverlay

    def test_feature_init(self):
        self.send_signal("Overlay Data", self.whitelight)
        self.assertEqual(self.widget.attr_value.name, "1.000000")

    def test_context_not_open_invalid(self):
        self.send_signal("Overlay Data", None)
        self.assertIsNone(self.widget.imageplot.attr_x)
        self.send_signal("Overlay Data", self.whitelight)
        self.assertIsNotNone(self.widget.imageplot.attr_x)

    def test_unknown(self):
        self.send_signal("Overlay Data", self.whitelight[:10])
        wait_for_image(self.widget)
        levels = self.widget.imageplot.img.levels
        self.send_signal("Overlay Data", self.whitelight_unknown[:10])
        wait_for_image(self.widget)
        levelsu = self.widget.imageplot.img.levels
        np.testing.assert_equal(levelsu, levels)

    def test_set_variable_color(self):
        data = self.whitelight
        self.send_signal("Overlay Data", data)
        with patch(
            "orangecontrib.spectroscopy.widgets.owhyper.ImageItemNan.setLookupTable"
        ) as p:
            self.widget.attr_value = data.domain["1.000000"]
            self.widget.imageplot.update_color_schema()
            self.widget._update_feature_value()
            wait_for_image(self.widget)
            np.testing.assert_equal(
                len(p.call_args[0][0]), 256
            )  # 256 for a continuous variable

    def test_add_visible_image(self):
        data = self.whitelight
        self.send_signal("Data", data)
        self.send_signal("Overlay Data", data)
        self.widget._attr_changed()
        self.widget._update_feature_value()

        out = self.get_output("Decorated Data")
        self.assertIsNotNone(out.attributes["visible_images"])

    def test_add_visible_image_empty(self):
        data = self.whitelight
        self.assertNotIn("visible_images", data.attributes)

        self.send_signal("Data", data)
        self.send_signal("Overlay Data", data)

        out = self.get_output("Decorated Data")
        self.assertNotIn(
            "visible_images", data.attributes
        )  # don't modify original data
        self.assertIsNotNone(out.attributes["visible_images"])

    def test_add_visible_image_existing(self):
        data = Orange.data.Table("agilent/4_noimage_agg256.dat")
        data.attributes["visible_images"] = TestVisibleImage.mock_visible_image_data()

        def names(data):
            return [im.name for im in data.attributes["visible_images"]]

        original_names = names(data)

        self.send_signal("Data", data)
        self.send_signal("Overlay Data", data)

        out = self.get_output("Decorated Data")
        self.assertEqual(original_names, names(data))
        self.assertEqual(original_names + ["Overlay Image"], names(out))

    def test_add_visible_image_samename(self):
        data = Orange.data.Table("agilent/4_noimage_agg256.dat")
        data.attributes["visible_images"] = TestVisibleImage.mock_visible_image_data()

        # Fake that one of the images is an overlay image
        data.attributes["visible_images"][1].name = "Overlay Image"

        def names(data):
            return [im.name for im in data.attributes["visible_images"]]

        original_names = names(data)

        self.send_signal("Data", data)
        self.send_signal("Overlay Data", data)

        out = self.get_output("Decorated Data")
        self.assertEqual(original_names, names(data))
        self.assertEqual(original_names + ["Overlay Image (1)"], names(out))


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
from AnyQt.QtCore import QItemSelectionModel, QItemSelection, QItemSelectionRange
import Orange
from Orange.data import ContinuousVariable, DiscreteVariable, Domain
from Orange.widgets.tests.base import WidgetTest
from orangecontrib.spectroscopy.widgets.owpolar import OWPolar

class TestOWPolar(WidgetTest):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.multifile = Orange.data.Table("polar/4-angle-ftir_multifile.tab")
        cls.in1 = Orange.data.Table("polar/4-angle-ftir_multiin1.tab")
        cls.in2 = Orange.data.Table("polar/4-angle-ftir_multiin2.tab")
        cls.in3 = Orange.data.Table("polar/4-angle-ftir_multiin3.tab")
        cls.in4 = Orange.data.Table("polar/4-angle-ftir_multiin4.tab")
        cls.multifile_polar = Orange.data.Table("polar/4-angle-ftir_multifile_polar-results.tab")
        cls.multifile_model = Orange.data.Table("polar/4-angle-ftir_multifile_model-results.tab")
        cls.multiin_polar = Orange.data.Table("polar/4-angle-ftir_multiin_polar-results.tab")
        cls.multiin_model = Orange.data.Table("polar/4-angle-ftir_multiin_model-results.tab")

    def setUp(self):
        self.widget = self.create_widget(OWPolar)

    def test_multifile_init(self):
        self.send_signal("Data", self.multifile, 0)

        testfeats = [ft for ft in self.multifile.domain.metas
                     if isinstance(ft, ContinuousVariable)]
        testfeats = testfeats + [ft for ft in self.multifile.domain.attributes
                     if isinstance(ft, ContinuousVariable)]
        polfeats = [ft for ft in self.widget.featureselect[:]
                    if isinstance(ft, ContinuousVariable)]
        self.assertEqual(polfeats, testfeats)
        testinputs = [inp for inp in self.multifile.domain
                      if isinstance(inp, DiscreteVariable)]
        self.assertEqual(self.widget.anglemetas[:], testinputs)
        testxy = [xy for xy in self.multifile.domain.metas
                  if isinstance(xy, ContinuousVariable)]
        self.assertEqual(self.widget.x_axis[:], testxy)
        self.assertEqual(self.widget.y_axis[:], testxy)

    def test_multifile_in(self):
        self.send_signal("Data", self.multifile, 0)

        self.assertTrue(self.widget.isEnabled())
        for i in self.widget.multiin_labels:
            self.assertFalse(i.isEnabled())
        for i in self.widget.multiin_lines:
            self.assertFalse(i.isEnabled())
        self.widget.angles = self.widget.anglemetas[0]
        self.assertEqual(self.widget.angles, self.multifile.domain.metas[2])
        self.widget._change_angles()
        self.assertEqual(len(self.widget.labels), 4)
        self.assertEqual(len(self.widget.lines), 4)
        self.assertEqual(self.widget.polangles, list(np.linspace(0, 180, 5)[:4]))
        for i in self.widget.labels:
            self.assertTrue(i.isEnabled())
        for i in self.widget.lines:
            self.assertTrue(i.isEnabled())
        self.widget.map_x = self.widget.x_axis[0]
        self.assertEqual(self.widget.map_x, self.multifile.domain.metas[0])
        self.widget.map_y = self.widget.y_axis[1]
        self.assertEqual(self.widget.map_y, self.multifile.domain.metas[1])
        self.widget.feats = [self.widget.feat_view.model()[:][2],
                                 self.widget.feat_view.model()[:][3]]
        self.assertEqual(self.widget.feats[0], self.multifile.domain.metas[3])
        self.assertEqual(self.widget.feats[1], self.multifile.domain.metas[4])
        self.widget.alphas = [0, 0]
        self.widget.invert_angles = True
        self.widget.autocommit = True
        self.commit_and_wait(self.widget, 20000)

        polar = self.get_output("Polar Data")
        model = self.get_output("Curve Fit model data")

        np.testing.assert_allclose(np.asarray(self.multifile_polar.metas, dtype=float),
                                   np.asarray(polar.metas, dtype=float), rtol=4e-06)
        np.testing.assert_allclose(np.asarray(self.multifile_polar.X, dtype=float),
                                   np.asarray(polar.X, dtype=float), rtol=5e-06)
        np.testing.assert_allclose(np.asarray(self.multifile_model.metas, dtype=float),
                                   np.asarray(model.metas, dtype=float), rtol=4e-06)
        np.testing.assert_allclose(np.asarray(self.multifile_model.X, dtype=float),
                                   np.asarray(model.X, dtype=float), rtol=5e-06)

    def test_multi_inputs(self):
        self.send_signal("Data", self.in1, 0, widget=self.widget)
        self.send_signal("Data", self.in2, 1, widget=self.widget)

        self.assertFalse(self.widget.anglesel.isEnabled())
        for i in self.widget.multiin_labels:
            self.assertFalse(i.isEnabled())
        for i in self.widget.multiin_lines:
            self.assertFalse(i.isEnabled())
        self.send_signal("Data", self.in3, 2, widget=self.widget)
        self.send_signal("Data", self.in4, 3, widget=self.widget)
        self.assertFalse(self.widget.anglesel.isEnabled())
        for i in self.widget.multiin_labels:
            self.assertTrue(i.isEnabled())
        for i in self.widget.multiin_lines:
            self.assertTrue(i.isEnabled())

        self.widget.map_x = self.widget.x_axis[0]
        self.assertEqual(self.widget.map_x, self.in1.domain.metas[0])
        self.widget.map_y = self.widget.y_axis[1]
        self.assertEqual(self.widget.map_y, self.in1.domain.metas[1])

        self.widget.feats = [self.widget.feat_view.model()[:][2],
                                 self.widget.feat_view.model()[:][3]]
        self.assertEqual(self.widget.feats[0],
                         self.in1.domain.metas[2].copy(compute_value=None))
        self.assertEqual(self.widget.feats[1],
                         self.in1.domain.metas[3].copy(compute_value=None))
        self.widget.alphas = [0, 0]
        self.widget.invert_angles = True
        self.widget.autocommit = True
        self.commit_and_wait(self.widget, 20000)

        polar = self.get_output("Polar Data")
        model = self.get_output("Curve Fit model data")

        np.testing.assert_allclose(
            np.asarray(self.multiin_polar.metas[:,np.r_[0:2,3:7]], dtype=float),
            np.asarray(polar.metas[:,np.r_[0:2,3:7]], dtype=float), rtol=2e-06)
        np.testing.assert_allclose(
            np.asarray(self.multiin_polar.metas[:,7:], dtype=float),
            np.asarray(polar.metas[:,7:], dtype=float), rtol=4e-06)
        np.testing.assert_allclose(self.multiin_polar.X, polar.X, rtol=5e-06)
        np.testing.assert_allclose(
            np.asarray(self.multiin_model.metas[:,np.r_[0:2,3:7]], dtype=float),
            np.asarray(model.metas[:,np.r_[0:2,3:7]], dtype=float), rtol=4e-06)
        np.testing.assert_allclose(
            np.asarray(self.multiin_model.metas[:,7:], dtype=float),
            np.asarray(model.metas[:,7:], dtype=float), rtol=4e-06)
        np.testing.assert_allclose(self.multiin_model.X, model.X, rtol=5e-06)

    def test_pixelsubset(self):
        #Test multi in with subset of pixels selected
        rng = np.random.default_rng()
        sub_idx = rng.choice(4, size=(2), replace=False)
        subset = self.in1[sub_idx]

        self.send_signal("Data", subset, 0, widget=self.widget)
        self.send_signal("Data", self.in2, 1, widget=self.widget)
        self.send_signal("Data", self.in3, 2, widget=self.widget)
        self.send_signal("Data", self.in4, 3, widget=self.widget)

        self.widget.map_x = self.widget.x_axis[0]
        self.widget.map_y = self.widget.y_axis[1]
        self.widget.feats = [self.widget.feat_view.model()[:][2],
                                 self.widget.feat_view.model()[:][3]]
        self.widget.alphas = [0, 0]
        self.widget.invert_angles = True
        self.widget.autocommit = True
        self.commit_and_wait(self.widget, 20000)

        polar = self.get_output("Polar Data")
        model = self.get_output("Curve Fit model data")

        self.assertEqual(len(polar), len(sub_idx)*4)
        self.assertEqual(len(model), len(sub_idx)*4)

    def test_multiin_mismatched_domain(self):

        metadom = self.in1.domain.metas
        metadom = [i for i in metadom if isinstance(i, ContinuousVariable)]
        attdom = self.in1.domain.attributes
        attdom = attdom[0::2]
        mismatched_domain = Domain(attdom, metas = metadom)
        mismatched_table = self.in1.transform(mismatched_domain)

        self.send_signal("Data", mismatched_table, 0, widget=self.widget)
        self.send_signal("Data", self.in2, 1, widget=self.widget)
        self.send_signal("Data", self.in3, 2, widget=self.widget)
        self.send_signal("Data", self.in4, 3, widget=self.widget)

        feat_len = len(metadom) + len(attdom) + 1
        XY_len = len(metadom)
        self.assertEqual(feat_len, len(self.widget.feat_view.model()[:]))
        self.assertEqual(XY_len, len(self.widget.x_axis[:]))
        self.assertEqual(XY_len, len(self.widget.y_axis[:]))

        self.send_signal("Data", self.in2, 0, widget=self.widget)
        self.send_signal("Data", self.in3, 1, widget=self.widget)
        self.send_signal("Data", mismatched_table, 2, widget=self.widget)
        self.send_signal("Data", self.in4, 3, widget=self.widget)

        feat_len = len(metadom) + len(attdom) + 1
        XY_len = len(metadom)
        self.assertEqual(feat_len, len(self.widget.feat_view.model()[:]))
        self.assertEqual(XY_len, len(self.widget.x_axis[:]))
        self.assertEqual(XY_len, len(self.widget.y_axis[:]))

    def test_custom_angles(self):
        # test inputting custom angles (multin and multifile)
        self.send_signal("Data", self.multifile, 0, widget=self.widget)
        angles = np.array([0, 22.5, 45.0, 90])

        for i, j in enumerate(self.widget.lines):
            j.setText(str(angles[i]))
        self.widget._send_angles()
        for i, j in enumerate(self.widget.polangles):
            self.assertEqual(j, angles[i])

        self.send_signal("Data", self.in1, 0, widget=self.widget)
        self.send_signal("Data", self.in2, 1, widget=self.widget)
        self.send_signal("Data", self.in3, 2, widget=self.widget)
        self.send_signal("Data", self.in4, 3, widget=self.widget)

        for i, j in enumerate(self.widget.multiin_lines):
            j.setText(str(angles[i]))
        self.widget._send_ind_angles()
        for i, j in enumerate(self.widget.polangles):
            self.assertEqual(j, angles[i])

    def test_warnings(self):
        #test all warnings
        self.send_signal("Data", self.multifile, 0, widget=self.widget)
        self.widget.autocommit = True

        self.commit_and_wait(self.widget)
        self.assertTrue(self.widget.Warning.nofeat.is_shown())

        self.widget.feats = [self.widget.feat_view.model()[:][4]]
        self.widget.map_x = None
        self.widget.map_y = None
        self.commit_and_wait(self.widget)
        self.assertTrue(self.widget.Warning.noxy.is_shown())

        self.widget.map_x = self.widget.x_axis[0]
        self.widget.map_y = self.widget.y_axis[1]
        self.widget.polangles = []
        self.commit_and_wait(self.widget)
        self.assertTrue(self.widget.Warning.pol.is_shown())
        self.widget.polangles = [0.0,45.0,'hi',135.0]
        self.commit_and_wait(self.widget)
        self.assertTrue(self.widget.Warning.pol.is_shown())

        self.widget.polangles = [0.0,45.0,90.0,135.0]
        self.widget.feats = [self.widget.feat_view.model()[:][0]]
        self.widget.alphas = [0]
        self.commit_and_wait(self.widget)
        self.assertTrue(self.widget.Warning.XYfeat.is_shown())

        self.send_signal("Data", self.in1, 0, widget=self.widget)
        self.send_signal("Data", self.in2, 1, widget=self.widget)
        self.assertTrue(self.widget.Warning.notenough.is_shown())

        self.send_signal("Data", self.in3, 2, widget=self.widget)
        self.assertTrue(self.widget.Warning.notenough.is_shown())

        self.send_signal("Data", self.in4, 3, widget=self.widget)
        self.assertFalse(self.widget.Warning.notenough.is_shown())

    def test_disconnect(self):
        self.send_signal("Data", self.multifile, 0, widget=self.widget)
        self.widget.angles = self.widget.anglemetas[0]
        self.widget.map_x = self.widget.x_axis[0]
        self.widget.map_y = self.widget.y_axis[1]
        self.widget.alphas = [0, 0]
        self.widget.invert_angles = True
        self.widget.autocommit = True
        self.widget.feats = [self.widget.feat_view.model()[:][2],
                            self.widget.feat_view.model()[:][3]]
        self.widget.handleNewSignals()
        self.wait_until_stop_blocking()
        self.send_signal("Data", None, 0, widget=self.widget)

    def test_alpha_changes(self):
        self.send_signal("Data", self.multifile, 0, widget=self.widget)
        self.widget.angles = self.widget.anglemetas[0]
        self.widget.map_x = self.widget.x_axis[0]
        self.widget.map_y = self.widget.y_axis[1]
        self.widget.alpha = 0
        vars = [self.widget.feat_view.model()[:][2],
                self.widget.feat_view.model()[:][3]]
        view = self.widget.feat_view
        model = self.widget.featureselect
        update_selection(model, view, vars)
        self.widget.change_alphas()
        self.assertEqual(self.widget.alphas, [0, 0])
        self.widget.alpha = 90
        update_selection(model, view, [vars[1]])
        self.widget.change_alphas()
        self.assertEqual(self.widget.alphas, [0, 90])
        update_selection(model, view, [vars[0]])
        self.widget.change_alphas()
        self.assertEqual(self.widget.alphas, [90, 90])

def update_selection(model, view, setting):
    selection = QItemSelection()
    sel_model = view.selectionModel()
    model_values = model[:]
    for var in setting:
        index = model_values.index(var)
        model_index = view.model().index(index, 0)
        selection.append(QItemSelectionRange(model_index))
    sel_model.select(selection, QItemSelectionModel.ClearAndSelect)
    # def test_clearangles(self):
    #     #test clearing angles
    #     pass

if __name__ == "__main__":
    unittest.main()



from AnyQt.QtCore import Qt
import scipy.sparse as sp

from Orange.widgets import gui
from Orange.widgets.widget import Msg
from Orange.data import Table, Domain, ContinuousVariable, StringVariable
from Orange.widgets.settings import Setting
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.signals import Output
from Orange.widgets.utils.widgetpreview import WidgetPreview

from orangecontrib.spectroscopy.models.pls import PLSRegressionLearner


class OWPLS(OWBaseLearner):
    name = 'PLS'
    description = "Partial Least Squares Regression widget for multivariate data analysis"
    icon = "icons/PLS.svg"
    keywords = ["partial least squares"]

    LEARNER = PLSRegressionLearner

    class Outputs(OWBaseLearner.Outputs):
        coefsdata = Output("Coefficients", Table, explicit=True)
        transformed = Output("Transformed Data", Table)
        components = Output("Components", Table)

    class Warning(OWBaseLearner.Warning):
        sparse_data = Msg('Sparse input data: default preprocessing is to scale it.')
        deprecated = Msg('The PLS widget is deprecated and will be removed in the future.\n'
                         'Please use the PLS widget from the Model category instead.')

    n_components = Setting(2)
    max_iter = Setting(500)

    def add_main_layout(self):
        self.Warning.deprecated()

        self.optimization_box = gui.vBox(
            self.controlArea, "Optimization Parameters")
        self.ncomps_spin = gui.spin(
            self.optimization_box, self, "n_components", 1, 50, 1,
            label="Components: ",
            alignment=Qt.AlignRight, controlWidth=100,
            callback=self.settings_changed)
        self.n_iters = gui.spin(
            self.optimization_box, self, "max_iter", 5, 1000000, 50,
            label="Iteration limit: ",
            alignment=Qt.AlignRight, controlWidth=100,
            callback=self.settings_changed,
            checkCallback=self.settings_changed)

    def update_model(self):
        super().update_model()
        coef_table = None
        projection = None
        components = None
        if self.model is not None:
            coef_table = self.model.coefficients_table()
            projection = self.model.project(self.data)
            components = self.model.components()
        self.Outputs.coefsdata.send(coef_table)
        self.Outputs.transformed.send(projection)
        self.Outputs.components.send(components)

    @OWBaseLearner.Inputs.data
    def set_data(self, data):
        # reimplemented completely because the base learner does not
        # allow multiclass

        self.Warning.sparse_data.clear()

        self.Error.data_error.clear()
        self.data = data

        if data is not None and data.domain.class_var is None and not data.domain.class_vars:
            self.Error.data_error(
                "Data has no target variable.\n"
                "Select one with the Select Columns widget.")
            self.data = None

        # invalidate the model so that handleNewSignals will update it
        self.model = None

        if self.data and sp.issparse(self.data.X):
            self.Warning.sparse_data()

    def create_learner(self):
        common_args = {'preprocessors': self.preprocessors}
        return PLSRegressionLearner(n_components=self.n_components,
                                    max_iter=self.max_iter,
                                    **common_args)


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWPLS).run(Table("housing"))



import numpy
import numpy as np

import Orange
from Orange.widgets.tests.base import WidgetTest
from orangecontrib.spectroscopy.data import getx
from orangecontrib.spectroscopy.widgets.owcos import OWCos, calc_cos, sort_data


class TestOWCOS(WidgetTest):
    # make test data
    WN = Orange.data.Domain([Orange.data.ContinuousVariable(str(w)) for w in [3, 1, 2]])
    X1 = np.array([[1, 2, 3], [4, 5, 6]])
    X2 = np.array([[2, 3, 4], [6, 7, 8]])

    DATA1 = Orange.data.Table.from_numpy(WN, X1)
    DATA2 = Orange.data.Table.from_numpy(WN, X2)

    def setUp(self):
        self.widget = self.create_widget(OWCos)

    def test_calc_cos(self):
        cos = calc_cos(self.DATA1, self.DATA2)
        numpy.testing.assert_array_equal(cos[0], [[6., 6., 6.],
                                                  [6., 8., 4.],
                                                  [6., 4., 8.]])

        numpy.testing.assert_array_almost_equal(cos[1], [[0., -0.95492966, 0.95492966],
                                                         [1.27323954, 0.31830989, 2.2281692],
                                                         [-1.27323954, -2.2281692, -0.31830989]])

        numpy.testing.assert_array_equal(cos[2], [[-1.5, -0.5, -2.5],
                                                  [1.5, 2.5, 0.5]])

        numpy.testing.assert_array_equal(cos[3], [[-2., -1., -3.],
                                                  [2., 3., 1.]])

        numpy.testing.assert_array_equal(cos[4], [1., 2., 3.])

        numpy.testing.assert_array_equal(cos[5], [1., 2., 3.])

    def test_sort_data(self):
        sorted_data = sort_data(self.DATA1)
        numpy.testing.assert_array_equal(sorted_data, [[2, 3, 1], [5, 6, 4]])
        numpy.testing.assert_array_equal(getx(sorted_data), [1, 2, 3])
